# Copyright 2016-2018 Berk Onat, Fawzi Mohamed
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
from .GromacsCommon import PARSERNAME, PROGRAMNAME, PARSERTAG
from nomadcore.smart_parser.SmartParserDictionary import metaNameConverter, metaNameConverter_UnderscoreSpaceDash
from nomadcore.smart_parser.SmartParserDictionary import MetaInfoMap, FileInfoMap, MapDictionary
from nomadcore.smart_parser.SmartParserDictionary import getDict_MetaStrInDict, getList_MetaStrInDict, get_unitDict
import nomadcore.md_data_access.MDDataAccess as MDData

def get_fileListDict():
    """Loads dictionary for file namelists.

    Returns:
        the list of defaults file namelists
    """

    startpage = {
        'nameTranslate'   : metaNameConverter,
        'metaHeader'      : PARSERTAG,
        'metaNameTag'     : 'inout_file',
        'metaInfoType'    : 'C',
        'activeMetaNames' : [],
        'activeSections'  : [PARSERTAG+'_section_input_output_files']
        }
    namelist = {
        'topoltpr'     : FileInfoMap(startpage, fileFormat=['.tpr', '.tpb'],
                                     fileInterface=['mdanalysis','pymolfile','mdtraj'],
                                     infoPurpose=['topology', 'unitcell'], matchNames="-s"),
        'trajtrr'      : FileInfoMap(startpage, fileFormat=['.trr', '.trj', '.xdr', '.cpt'],
                                     #fileInterface=['mdanalysis','pymolfile','mdtraj'],
                                     infoPurpose=['trajectory', 'unitcell'], matchNames="-o"),
        'traj_compxtc' : FileInfoMap(startpage, fileFormat=['.xtc', '.xdr'],
                                     #fileInterface=['mdanalysis','pymolfile','mdtraj'],
                                     infoPurpose=['trajectory', 'unitcell'], matchNames="-x"),
        #'statecpt'     : FileInfoMap(startpage, activeInfo=True, fileFormat=['.cpt'],
        #                             fileInterface=['mdanalysis','mdtraj'],
        #                             infoPurpose=['inputcoordinates'],
        #                             matchNames="-cpi"),
        #'confoutgro'   : FileInfoMap(startpage, activeInfo=True, fileFormat=['.gro', '.g96', '.pdb'],
        #                             fileInterface=['mdanalysis','mdtraj'],
        #                             infoPurpose=['inputcoordinates'],
        #                             matchNames="-c"),
        'eneredr'      : FileInfoMap(startpage, activeInfo=True, fileFormat=['.edr'],
                                     infoPurpose=['thermostats'], matchNames="-e"),
        'mdlog'        : FileInfoMap(startpage),
        'rerunxtc'     : FileInfoMap(startpage),
        'rotationlog'  : FileInfoMap(startpage),
        'rotangleslog' : FileInfoMap(startpage),
        'rotslabslog'  : FileInfoMap(startpage),
        'rottorquelog' : FileInfoMap(startpage),
        'nmmtx'        : FileInfoMap(startpage),
        'dipolendx'    : FileInfoMap(startpage),
        'membeddat'    : FileInfoMap(startpage),
        'membedtop'    : FileInfoMap(startpage),
        'membedndx'    : FileInfoMap(startpage)
        }
    return MapDictionary(namelist)

def get_nameListDict(self, deflist):
    """Loads control in data of GROMACS.

    Args:
        deflist: name list definition (cntrl/arch/parm/qmopts/grpopts/anneal).

    matchWith parameters:
        EOL = End of line
        PL  = Match with previous line
        PW  = Previous word
        NW  = Next word (until space, comma) (DEFAULT)
        UD  = until delimeter (can be any string/character)

        The text matched inside apostrophe or qoutation is extracted

    Returns:
        the list of namelists
    """
    startpage = {
        'nameTranslate'   : metaNameConverter_UnderscoreSpaceDash,
        'metaHeader'      : PARSERTAG,
        'metaNameTag'     : 'inout_control',
        'metaInfoType'    : 'C',
        'activeMetaNames' : [],
        'activeSections'  : [PARSERTAG+'_section_control_parameters'],
        'matchWith'       : 'EOL'
        }

    archlist = {
        'GROMACS version'    : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'Precision'          : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'Memory model'       : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'MPI library'        : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'OpenMP support'     : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'GPU support'        : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'OpenCL support'     : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'invsqrt routine'    : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'SIMD instructions'  : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'FFT library'        : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'RDTSCP usage'       : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'C[+][+]11 compilation'  : MetaInfoMap(startpage, defaultValue='',
                                               replaceTag='Cxx11 compilation', matchWith='EOL'),
        'TNG support'        : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'Tracing support'    : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'Built on'           : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'Built by'           : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'Build OS/arch'      : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'Build CPU vendor'   : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'Build CPU brand'    : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'Build CPU family'   : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'Build CPU features' : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'C compiler'         : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'C compiler flags'   : MetaInfoMap(startpage, defaultValue='', matchWith='EOL'),
        'C[+][+] compiler'       : MetaInfoMap(startpage, defaultValue='',
                                               replaceTag='Cxx compiler', matchWith='EOL'),
        'C[+][+] compiler flags' : MetaInfoMap(startpage, defaultValue='',
                                               replaceTag='Cxx compiler flags', matchWith='EOL'),
        'Boost version'      : MetaInfoMap(startpage, defaultValue='', matchWith='EOL')
        }

    startpage = {
        'nameTranslate'   : metaNameConverter,
        'metaHeader'      : PARSERTAG,
        'metaNameTag'     : 'inout_control',
        'metaInfoType'    : 'C',
        'activeMetaNames' : [],
        'activeSections'  : [PARSERTAG+'_section_control_parameters'],
        'matchWith'       : 'EOL'
        }

    cntrllist = {
        'integrator'         : MetaInfoMap(startpage, defaultValue=''),
        'tinit'              : MetaInfoMap(startpage, defaultValue=0),
        'dt'                 : MetaInfoMap(startpage, defaultValue=0.0001),
        'nsteps'             : MetaInfoMap(startpage, defaultValue=0),
        'init-step'          : MetaInfoMap(startpage, defaultValue=0),
        'simulation-part'    : MetaInfoMap(startpage, defaultValue=0),
        'comm-mode'          : MetaInfoMap(startpage, defaultValue=0),
        'nstcomm'            : MetaInfoMap(startpage, defaultValue=0),
        'bd-fric'            : MetaInfoMap(startpage, defaultValue=0),
        'ld-seed'            : MetaInfoMap(startpage, defaultValue=0),
        'emtol'              : MetaInfoMap(startpage, defaultValue=0),
        'emstep'             : MetaInfoMap(startpage, defaultValue=0),
        'niter'              : MetaInfoMap(startpage, defaultValue=0),
        'fcstep'             : MetaInfoMap(startpage, defaultValue=0),
        'nstcgsteep'         : MetaInfoMap(startpage, defaultValue=0),
        'nbfgscorr'          : MetaInfoMap(startpage, defaultValue=0),
        'rtpi'               : MetaInfoMap(startpage, defaultValue=0),
        'nstxout'            : MetaInfoMap(startpage, defaultValue=0),
        'nstvout'            : MetaInfoMap(startpage, defaultValue=0),
        'nstfout'            : MetaInfoMap(startpage, defaultValue=0),
        'nstlog'             : MetaInfoMap(startpage, defaultValue=0),
        'nstcalcenergy'      : MetaInfoMap(startpage, defaultValue=0),
        'nstenergy'          : MetaInfoMap(startpage, defaultValue=0),
        'nstxout-compressed' : MetaInfoMap(startpage, defaultValue=0),
        'compressed-x-precision' : MetaInfoMap(startpage, defaultValue=0),
        'cutoff-scheme'      : MetaInfoMap(startpage, defaultValue=0),
        'nstlist'            : MetaInfoMap(startpage, defaultValue=0),
        'ns-type'            : MetaInfoMap(startpage, defaultValue=0),
        'pbc'                : MetaInfoMap(startpage, defaultValue=0),
        'periodic-molecules' : MetaInfoMap(startpage, defaultValue=0),
        'verlet-buffer-tolerance': MetaInfoMap(startpage, defaultValue=0),
        'rlist'              : MetaInfoMap(startpage, defaultValue=0),
        'rlistlong'          : MetaInfoMap(startpage, defaultValue=0),
        'nstcalclr'          : MetaInfoMap(startpage, defaultValue=0),
        'coulombtype'        : MetaInfoMap(startpage, defaultValue=0),
        'coulomb-modifier'   : MetaInfoMap(startpage, defaultValue=0),
        'rcoulomb-switch'    : MetaInfoMap(startpage, defaultValue=0),
        'rcoulomb'           : MetaInfoMap(startpage, defaultValue=0),
        'epsilon-r'          : MetaInfoMap(startpage, defaultValue=0),
        'epsilon-rf'         : MetaInfoMap(startpage, defaultValue=0),
        'vdw-type'           : MetaInfoMap(startpage, defaultValue=0),
        'vdw-modifier'       : MetaInfoMap(startpage, defaultValue=0),
        'rvdw-switch'        : MetaInfoMap(startpage, defaultValue=0),
        'rvdw'               : MetaInfoMap(startpage, defaultValue=0),
        'DispCorr'           : MetaInfoMap(startpage, defaultValue=0),
        'table-extension'    : MetaInfoMap(startpage, defaultValue=0),
        'fourierspacing'     : MetaInfoMap(startpage, defaultValue=0),
        'fourier-nx'         : MetaInfoMap(startpage, defaultValue=0),
        'fourier-ny'         : MetaInfoMap(startpage, defaultValue=0),
        'fourier-nz'         : MetaInfoMap(startpage, defaultValue=0),
        'pme-order'          : MetaInfoMap(startpage, defaultValue=0),
        'ewald-rtol'         : MetaInfoMap(startpage, defaultValue=0),
        'ewald-rtol-lj'      : MetaInfoMap(startpage, defaultValue=0),
        'lj-pme-comb-rule'   : MetaInfoMap(startpage, defaultValue=0),
        'ewald-geometry'     : MetaInfoMap(startpage, defaultValue=0),
        'epsilon-surface'    : MetaInfoMap(startpage, defaultValue=0),
        'implicit-solvent'   : MetaInfoMap(startpage, defaultValue=0),
        'gb-algorithm'       : MetaInfoMap(startpage, defaultValue=0),
        'nstgbradii'         : MetaInfoMap(startpage, defaultValue=0),
        'rgbradii'           : MetaInfoMap(startpage, defaultValue=0),
        'gb-epsilon-solvent' : MetaInfoMap(startpage, defaultValue=0),
        'gb-saltconc'        : MetaInfoMap(startpage, defaultValue=0),
        'gb-obc-alpha'       : MetaInfoMap(startpage, defaultValue=0),
        'gb-obc-beta'        : MetaInfoMap(startpage, defaultValue=0),
        'gb-obc-gamma'       : MetaInfoMap(startpage, defaultValue=0),
        'gb-dielectric-offset' : MetaInfoMap(startpage, defaultValue=0),
        'sa-algorithm'       : MetaInfoMap(startpage, defaultValue=0),
        'sa-surface-tension' : MetaInfoMap(startpage, defaultValue=0),
        'tcoupl'             : MetaInfoMap(startpage, defaultValue=0),
        'nsttcouple'         : MetaInfoMap(startpage, defaultValue=0),
        'nh-chain-length'    : MetaInfoMap(startpage, defaultValue=0),
        'print-nose-hoover-chain-variables' : MetaInfoMap(startpage, defaultValue=0),
        'pcoupl'             : MetaInfoMap(startpage, defaultValue=0),
        'pcoupltype'         : MetaInfoMap(startpage, defaultValue=0),
        'nstpcouple'         : MetaInfoMap(startpage, defaultValue=0),
        'tau-p'              : MetaInfoMap(startpage, defaultValue=0),
        'compressibility\[    0\]' : MetaInfoMap(startpage, defaultValue=0,
                replaceDict = {'{' : '[', '}' : ']'},
                subFunc=lambda x: np.asarray(
                    self.literal_eval(x)) if x is not None else None),
        'compressibility\[    1\]' : MetaInfoMap(startpage, defaultValue=0,
                replaceDict = {'{' : '[', '}' : ']'},
                subFunc=lambda x: np.asarray(
                    self.literal_eval(x)) if x is not None else None),
        'compressibility\[    2\]' : MetaInfoMap(startpage, defaultValue=0,
                replaceDict = {'{' : '[', '}' : ']'},
                subFunc=lambda x: np.asarray(
                    self.literal_eval(x)) if x is not None else None),
        'ref-p\[    0\]'       : MetaInfoMap(startpage, defaultValue=0,
                replaceDict = {'{' : '[', '}' : ']'},
                subFunc=lambda x: np.asarray(
                    self.literal_eval(x)) if x is not None else None),
        'ref-p\[    1\]'       : MetaInfoMap(startpage, defaultValue=0,
                replaceDict = {'{' : '[', '}' : ']'},
                subFunc=lambda x: np.asarray(
                    self.literal_eval(x)) if x is not None else None),
        'ref-p\[    2\]'       : MetaInfoMap(startpage, defaultValue=0,
                replaceDict = {'{' : '[', '}' : ']'},
                subFunc=lambda x: np.asarray(
                    self.literal_eval(x)) if x is not None else None),
        'refcoord-scaling'   : MetaInfoMap(startpage, defaultValue=0),
        'posres-com\[0\]'      : MetaInfoMap(startpage, defaultValue=0),
        'posres-com\[1\]'      : MetaInfoMap(startpage, defaultValue=0),
        'posres-com\[2\]'      : MetaInfoMap(startpage, defaultValue=0),
        'posres-comB\[0\]'     : MetaInfoMap(startpage, defaultValue=0),
        'posres-comB\[1\]'     : MetaInfoMap(startpage, defaultValue=0),
        'posres-comB\[2\]'     : MetaInfoMap(startpage, defaultValue=0),
        'QMMM'               : MetaInfoMap(startpage, defaultValue=0),
        'QMconstraints'      : MetaInfoMap(startpage, defaultValue=0),
        'QMMMscheme'         : MetaInfoMap(startpage, defaultValue=0),
        'MMChargeScaleFactor' : MetaInfoMap(startpage, defaultValue=0),
        }

    qmlist = {
        'ngQM' : MetaInfoMap(startpage, defaultValue=0),
        'constraint-algorithm' : MetaInfoMap(startpage, defaultValue=0),
        'continuation' : MetaInfoMap(startpage, defaultValue=0),
        'Shake-SOR' : MetaInfoMap(startpage, defaultValue=0),
        'shake-tol' : MetaInfoMap(startpage, defaultValue=0),
        'lincs-order' : MetaInfoMap(startpage, defaultValue=0),
        'lincs-iter' : MetaInfoMap(startpage, defaultValue=0),
        'lincs-warnangle' : MetaInfoMap(startpage, defaultValue=0),
        'nwall' : MetaInfoMap(startpage, defaultValue=0),
        'wall-type' : MetaInfoMap(startpage, defaultValue=0),
        'wall-r-linpot' : MetaInfoMap(startpage, defaultValue=0),
        'wall-atomtype\[0\]' : MetaInfoMap(startpage, defaultValue=0),
        'wall-atomtype\[1\]' : MetaInfoMap(startpage, defaultValue=0),
        'wall-density\[0\]' : MetaInfoMap(startpage, defaultValue=0),
        'wall-density\[1\]' : MetaInfoMap(startpage, defaultValue=0),
        'wall-ewald-zfac' : MetaInfoMap(startpage, defaultValue=0),
        'pull' : MetaInfoMap(startpage, defaultValue=0),
        'rotation' : MetaInfoMap(startpage, defaultValue=0),
        'interactiveMD' : MetaInfoMap(startpage, defaultValue=0),
        'disre' : MetaInfoMap(startpage, defaultValue=0),
        'disre-weighting' : MetaInfoMap(startpage, defaultValue=0),
        'disre-mixed' : MetaInfoMap(startpage, defaultValue=0),
        'dr-fc' : MetaInfoMap(startpage, defaultValue=0),
        'dr-tau' : MetaInfoMap(startpage, defaultValue=0),
        'nstdisreout' : MetaInfoMap(startpage, defaultValue=0),
        'orire-fc' : MetaInfoMap(startpage, defaultValue=0),
        'orire-tau' : MetaInfoMap(startpage, defaultValue=0),
        'nstorireout' : MetaInfoMap(startpage, defaultValue=0),
        'free-energy' : MetaInfoMap(startpage, defaultValue=0),
        'cos-acceleration' : MetaInfoMap(startpage, defaultValue=0),
        'deform\[    0\]' : MetaInfoMap(startpage, defaultValue=0,
                replaceDict = {'{' : '[', '}' : ']'},
                subFunc=lambda x: np.asarray(
                    self.literal_eval(x)) if x is not None else None),
        'deform\[    1\]' : MetaInfoMap(startpage, defaultValue=0,
                replaceDict = {'{' : '[', '}' : ']'},
                subFunc=lambda x: np.asarray(
                    self.literal_eval(x)) if x is not None else None),
        'deform\[    2\]' : MetaInfoMap(startpage, defaultValue=0,
                replaceDict = {'{' : '[', '}' : ']'},
                subFunc=lambda x: np.asarray(
                    self.literal_eval(x)) if x is not None else None),
        'simulated-tempering' : MetaInfoMap(startpage, defaultValue=0),
        'E-x' : MetaInfoMap(startpage, defaultValue=0, nextMatch='n', concatMatch=True),
        'E-xt' : MetaInfoMap(startpage, defaultValue=0, nextMatch='n', concatMatch=True),
        'E-y' : MetaInfoMap(startpage, defaultValue=0, nextMatch='n', concatMatch=True),
        'E-yt' : MetaInfoMap(startpage, defaultValue=0, nextMatch='n', concatMatch=True),
        'E-z' : MetaInfoMap(startpage, defaultValue=0, nextMatch='n', concatMatch=True),
        'E-zt' : MetaInfoMap(startpage, defaultValue=0, nextMatch='n', concatMatch=True),
        'swapcoords' : MetaInfoMap(startpage, defaultValue=0),
        'adress' : MetaInfoMap(startpage, defaultValue=0),
        'userint1' : MetaInfoMap(startpage, defaultValue=0),
        'userint2' : MetaInfoMap(startpage, defaultValue=0),
        'userint3' : MetaInfoMap(startpage, defaultValue=0),
        'userint4' : MetaInfoMap(startpage, defaultValue=0),
        'userreal1' : MetaInfoMap(startpage, defaultValue=0),
        'userreal2' : MetaInfoMap(startpage, defaultValue=0),
        'userreal3' : MetaInfoMap(startpage, defaultValue=0),
        'userreal4' : MetaInfoMap(startpage, defaultValue=0),
        }

    grplist = {
        'nrdf' : MetaInfoMap(startpage, defaultValue=0),
        'ref-t' : MetaInfoMap(startpage, defaultValue=0),
        'tau-t' : MetaInfoMap(startpage, defaultValue=0),
        }

    anneallist = {
        'annealing' : MetaInfoMap(startpage, defaultValue=0),
        'annealing-npoints' : MetaInfoMap(startpage, defaultValue=0),
        'acc' : MetaInfoMap(startpage, defaultValue=0,
            subFunc=lambda x: [float(i) for i in x.split()] if(x is not None and
                isinstance(x,str) is True) else None),
        'nfreeze' : MetaInfoMap(startpage, defaultValue=0,
            subFunc=lambda x: x.split() if(x is not None and
                isinstance(x,str) is True) else None),
        'energygrp-flags\[  0\]' : MetaInfoMap(startpage, defaultValue=0),
        'energygrp-flags\[  1\]' : MetaInfoMap(startpage, defaultValue=0),
        'energygrp-flags\[  2\]' : MetaInfoMap(startpage, defaultValue=0),
        }

    startpage.update({
        'nameTranslate'   : metaNameConverter_UnderscoreSpaceDash,
        'metaNameTag'     : 'parm'
        })
    parmlist = {
        'Domain Decomposition on' : MetaInfoMap(startpage, defaultValue=0, replaceTag='Domain Decomposition Ranks', matchWith='EOL'),
        'Dynamic load balancing' : MetaInfoMap(startpage, defaultValue=0),
        '\(option -dds\)' : MetaInfoMap(startpage, defaultValue=0, matchWith='NW'),
        'Optimizing the DD grid for' : MetaInfoMap(startpage, defaultValue=0, replaceTag='Number of cells'),
        'with a minimum initial size of' : MetaInfoMap(startpage, defaultValue=0, replaceTag='Min init size cells', matchWith='EOL'),
        'Domain decomposition grid' : MetaInfoMap(startpage, defaultValue=0),
        'separate PME ranks' : MetaInfoMap(startpage, defaultValue=0),
        'Domain decomposition rank' : MetaInfoMap(startpage, defaultValue=0),
        'coordinates' : MetaInfoMap(startpage, defaultValue=0, replaceTag='Domain Decomposition coordinates'),
        'Table routines are used for coulomb' : MetaInfoMap(startpage, defaultValue=0, replaceTag='use coulomb tables'),
        'Table routines are used for vdw' : MetaInfoMap(startpage, defaultValue=0, replaceTag='use vdw tables'),
        'NS' : MetaInfoMap(startpage, defaultValue=0, alsoMatch=["Cut-off's:"], replaceTag='cutoff ns'),
        'Coulomb' : MetaInfoMap(startpage, defaultValue=0, alsoMatch=["Cut-off's:"], replaceTag='cutoff coulomb'),
        'LJ' : MetaInfoMap(startpage, defaultValue=0, alsoMatch=["Cut-off's:"], replaceTag='cutoff lj'),
        'System total charge' : MetaInfoMap(startpage, defaultValue=0),
        'data points for 1-4 COUL' : MetaInfoMap(startpage, defaultValue=0, matchWith='PW'),
        'data points for 1-4 LJ6' : MetaInfoMap(startpage, defaultValue=0, matchWith='PW'),
        'data points for 1-4 LJ12' : MetaInfoMap(startpage, defaultValue=0, matchWith='PW'),
        'Potential shift: LJ r\^-12:' : MetaInfoMap(startpage, defaultValue=0, replaceTag='Potential shift LJ r12'),
        'r\^-6:' : MetaInfoMap(startpage, defaultValue=0, replaceTag='Potential shift LJ r6'),
        ', Coulomb' : MetaInfoMap(startpage, defaultValue=0, replaceTag='Potential shift Coulomb'),
        'Initializing Parallel LINear Constraint' : MetaInfoMap(startpage, defaultValue='yes', changeTags={'PLINCS' : 'Yes'}, mute=True),
        'PLINCS' : MetaInfoMap(startpage, ignoreMatch=True),
        'number of constraints is' : MetaInfoMap(startpage, defaultValue=0, removeText=' is'),
        'domain decomposition cell size is' : MetaInfoMap(startpage, defaultValue=0, replaceTag='Min size cells'),
        'shrink of DD cells \(option -dds\) is' : MetaInfoMap(startpage, defaultValue=0, removeText=' \(option -dds\) is'),
        'allowed shrink of domain decomposition cells is' : MetaInfoMap(startpage, defaultValue=0, removeText=' is'),
        'non-bonded interactions' : MetaInfoMap(startpage, defaultValue=0),
        'two-body bonded interactions  \(-rdd\)' : MetaInfoMap(startpage, defaultValue=0, replaceTag='two-body bonded interactions'),
        'multi-body bonded interactions  \(-rdd\)' : MetaInfoMap(startpage, defaultValue=0, replaceTag='multi-body bonded interactions'),
        'atoms separated by up to 4 constraints  \(-rcon\)' : MetaInfoMap(startpage, defaultValue=0, replaceTag='atoms separated 4 constraints'),
        'domain decomposition grid' : MetaInfoMap(startpage, defaultValue=0),
        'Center of mass motion removal mode is' : MetaInfoMap(startpage, defaultValue=0, removeText=' is'),
        'Atom distribution over' : MetaInfoMap(startpage, defaultValue=0, replaceTag='Atom distribution domains'),
        'Grid' : MetaInfoMap(startpage, defaultValue=0),
        'Initial temperature' : MetaInfoMap(startpage, defaultValue=0),
        }

    startpage.update({
        'metaNameTag'     : 'mdout',
        'activeSections'  : ['section_single_configuration_calculation']
        })
    mddatalist = {
        'Step' : MetaInfoMap(startpage),
        'Time' : MetaInfoMap(startpage),
        'Lambda' : MetaInfoMap(startpage),
        'Bond' : MetaInfoMap(startpage),
        'Angle' : MetaInfoMap(startpage),
        'G96Bond' : MetaInfoMap(startpage),
        'G96Angle' : MetaInfoMap(startpage),
        'U-B' : MetaInfoMap(startpage),
        'Proper Dih.' : MetaInfoMap(startpage),
        'Improper Dih.' : MetaInfoMap(startpage),
        'CMAP Dih.' : MetaInfoMap(startpage),
        'LJ (SR)' : MetaInfoMap(startpage),
        'LJ (LR)' : MetaInfoMap(startpage),
        'Disper. corr.' : MetaInfoMap(startpage),
        'Coulomb (SR)' : MetaInfoMap(startpage),
        'Coul. recip.' : MetaInfoMap(startpage),
        'LJ-14' : MetaInfoMap(startpage),
        'Coulomb-14' : MetaInfoMap(startpage),
        'Potential' : MetaInfoMap(startpage),
        'Kinetic En.' : MetaInfoMap(startpage),
        'Total Energy' : MetaInfoMap(startpage),
        'Conserved En.' : MetaInfoMap(startpage),
        'Temperature' : MetaInfoMap(startpage),
        'Pres. DC (bar)' : MetaInfoMap(startpage),
        'Pressure (bar)' : MetaInfoMap(startpage),
        'Constr. rmsd' : MetaInfoMap(startpage),
        'dVremain/dl' : MetaInfoMap(startpage),
        'Virial-Tensor' : MetaInfoMap(startpage),
        'Pressure-Tensor' : MetaInfoMap(startpage),
        }

    startpage.update({
        'nameTranslate'   : metaNameConverter_UnderscoreSpaceDash,
        'metaNameTag'     : 'mdout',
        'activeSections'  : ['section_single_configuration_calculation']
        })
    edrdatalist = {
        'Time' : MetaInfoMap(startpage),
        'Bond' : MetaInfoMap(startpage),
        'Angle' : MetaInfoMap(startpage),
        'G96Bond' : MetaInfoMap(startpage),
        'G96Angle' : MetaInfoMap(startpage),
        'U-B' : MetaInfoMap(startpage),
        'Proper Dih.' : MetaInfoMap(startpage),
        'Improper Dih.' : MetaInfoMap(startpage),
        'CMAP Dih.' : MetaInfoMap(startpage),
        'LJ (SR)' : MetaInfoMap(startpage),
        'LJ (LR)' : MetaInfoMap(startpage),
        'Disper. corr.' : MetaInfoMap(startpage),
        'Coulomb (SR)' : MetaInfoMap(startpage),
        'Coul. recip.' : MetaInfoMap(startpage),
        'LJ-14' : MetaInfoMap(startpage),
        'Coulomb-14' : MetaInfoMap(startpage),
        'Potential' : MetaInfoMap(startpage),
        'Kinetic En.' : MetaInfoMap(startpage),
        'Total Energy' : MetaInfoMap(startpage),
        'Conserved En.' : MetaInfoMap(startpage),
        'Temperature' : MetaInfoMap(startpage),
        'Pres. DC (bar)' : MetaInfoMap(startpage),
        'Pressure (bar)' : MetaInfoMap(startpage),
        'Constr. rmsd' : MetaInfoMap(startpage),
        'dVremain/dl' : MetaInfoMap(startpage),
        'Virial-Tensor' : MetaInfoMap(startpage),
        'Pressure-Tensor' : MetaInfoMap(startpage),
        }

    startpage.update({
        'metaNameTag'     : 'parm',
        'activeSections'  : [PARSERTAG+'_mdin_method']
        })
    extralist = {
        'flags' : MetaInfoMap(startpage),
        'number_of_atoms' : MetaInfoMap(startpage),
        'box_info' : MetaInfoMap(startpage),
        'unitcell_radius' : MetaInfoMap(startpage),
        'total_memory' : MetaInfoMap(startpage),
        'file_format' : MetaInfoMap(startpage),
        'file_version' : MetaInfoMap(startpage),
        'file_date' : MetaInfoMap(startpage),
        'file_time' : MetaInfoMap(startpage)
        }

    startpage.update({
        'metaNameTag'     : 'mdout',
        'activeSections'  : ['section_single_configuration_calculation']
        })
    # Additional dictionary to store stepping parameters
    # such as output log frequency, trajectory writing
    # freuquency and the actually needed number of
    # section_single_configuration_calculation
    #stepcontrol = {
    #    'logsteps'       : MetaInfoMap(startpage),
    #    'trajsteps'      : MetaInfoMap(startpage),
    #    'enersteps'      : MetaInfoMap(startpage),
    #    'steps'          : MetaInfoMap(startpage),
    #    'follow'         : MetaInfoMap(startpage),
    #    'sectioncontrol' : MetaInfoMap(startpage)
    #    }

    if deflist == 'mddata':
        namelist = mddatalist
    elif deflist == 'edrdata':
        namelist = edrdatalist
    elif deflist == 'arch':
        namelist = archlist
    elif deflist == 'qm':
        namelist = qmlist
    elif deflist == 'grp':
        namelist = grplist
    elif deflist == 'anneal':
        namelist = anneallist
    elif deflist == 'parm':
        namelist = parmlist
    elif deflist == 'extra':
        namelist = extralist
    elif deflist == 'stepcontrol':
        namelist = stepcontrol
    else:
        namelist = cntrllist
    return MapDictionary(namelist)

def set_Dictionaries(self):
        self.unitDict = get_unitDict('si')
        self.fileDict = get_fileListDict()
        self.cntrlDict = get_nameListDict(self, 'cntrl')
        self.archDict = get_nameListDict(self, 'arch')
        self.qmDict = get_nameListDict(self, 'qm')
        self.grpDict = get_nameListDict(self, 'grp')
        self.edrDict = get_nameListDict(self, 'edrdata')
        self.annealDict = get_nameListDict(self, 'anneal')
        self.parmDict = get_nameListDict(self, 'parm')
        self.mddataDict = get_nameListDict(self, 'mddata')
        self.extraDict = get_nameListDict(self, 'extra')
        #self.stepcontrolDict = get_nameListDict('stepcontrol')
        self.stepcontrolDict = {
            'logsteps'       : None,
            'nextlogsteps'   : None,
            'trajsteps'      : None,
            'enersteps'      : None,
            'steps'          : None,
            'follow'         : None,
            'sectioncontrol' : None,
        }
        self.metaDicts = {
            'file'   : self.fileDict,
            'cntrl'  : self.cntrlDict,
            'arch'   : self.archDict,
            'qm'     : self.qmDict,
            'grp'    : self.grpDict,
            'anneal' : self.annealDict,
            'parm'   : self.parmDict,
            'mddata' : self.mddataDict,
            'edr'    : self.edrDict,
            'extra'  : self.extraDict,
            }
        self.sectionDict = {
                'section' : {
                    "metaNameTag"      : ['input_output_files',
                                          'control_parameters'],
                    "listTypStr"       : 'type_section',
                    "repeatingSection" : False,
                    "supraNames"       : ["section_run"]
                    }
                }


def get_updateDictionary(self, defname):

    def value_convert_array(itemdict):
        val = itemdict["value"]
        if(val is not None and isinstance(val, str)):
            if("[" in val or "(" in val):
                return False, np.asarray(val), itemdict
        return False, val, itemdict

    def value_convert_array_norm(itemdict):
        val = []
        depdict = itemdict["depends"]
        for item in depdict["value"]:
            st=self.metaStorage.fetchAttrValue(item)
            if(st is not None and isinstance(st, str)):
                if("[" in st or "(" in st):
                    val.append(list(storedtext))
        if val:
            return False, np.linalg.norm(np.asarray(val)), itemdict
        else:
            return False, itemdict["value"], itemdict

    startpage = {
        'metaHeader'      : '',
        'metaNameTag'     : '',
        'activeMetaNames' : [],
        'activeSections'  : ['section_sampling_method','section_single_configuration_calculation',PARSERTAG+'_mdout']
        }

    # ---------------------------------------------------------------
    #   Definitions of meta data values for section_sampling_method
    # ---------------------------------------------------------------
    sampling = {
        'ensemble_type' : MetaInfoMap(startpage, activeInfo=True,
            depends=[
                {'test' : [['integrator', ' in [\"steep\",'
                                          '\"cg\",'
                                          '\"l-bfgs\"]']],
                 'assign' : 'minimization'},
                {'test' : [['integrator', ' in [\"md\",'
                                          '\"md-vv\",'
                                          '\"sd\",'
                                          '\"sd2\",'
                                          '\"md-vv-avek\"]'],
                           ['tcoupl', ' in [\"no\",\"No\",\"NO\"]']],
                 'assign' : 'NVE'},
                {'test' : [['integrator', ' in [\"md\",'
                                          '\"sd\",'
                                          '\"sd2\",'
                                          '\"md-vv\",'
                                          '\"md-vv-avek\"]'],
                           ['tcouple', ' in [\"berendsen\",'
                                       '\"nose-hoover\",'
                                       '\"andersen\",'
                                       '\"anderson-massive\",'
                                       '\"v-rescale\"]'],
                           ['pcoupl', ' in [\"no\",\"No\",\"NO\"]']],
                 'assign' : 'NVT'},
                {'test' : [['integrator', ' in [\"md\",'
                                          '\"sd\",'
                                          '\"sd2\",'
                                          '\"md-vv\",'
                                          '\"md-vv-avek\"]'],
                           ['tcouple', ' in [\"berendsen\",'
                                       '\"nose-hoover\",'
                                       '\"andersen\",'
                                       '\"anderson-massive\",'
                                       '\"v-rescale\"]'],
                           ['pcoupl', ' in [\"berendsen\",'
                                      '\"Parinello-Rahman\"]']],
                 'assign' : 'NPT'},
                {'test' : [['integrator', '== \"bd\"']],
                 'assign' : 'Langevin'}
                ],
            lookupdict=self.cntrlDict
            ),
        'sampling_method' : MetaInfoMap(startpage, activeInfo=True,
            depends=[
                {'test' : [['integrator', ' in [\"steep\",'
                                          '\"cg\",'
                                          '\"l-bfgs\"]']],
                 'assign' : 'geometry_optimization'},
                {'test' : [['integrator', ' in [\"md\",'
                                          '\"sd\",'
                                          '\"sd2\",'
                                          '\"bd\",'
                                          '\"md-vv\",'
                                          '\"md-vv-avek\"]']],
                 'assign' : 'molecular_dynamics'}
                ],
            lookupdict=self.cntrlDict
            ),
#            'settings_geometry_optimization' : MetaInfoMap(startpage),
#            'settings_metadynamics' : MetaInfoMap(startpage),
#            'settings_molecular_dynamics' : MetaInfoMap(startpage),
#            'settings_Monte_Carlo' : MetaInfoMap(startpage),
#        'geometry_optimization_energy_change' : MetaInfoMap(startpage,
#            depends={
#                '' : {'imin' : '1'},
#                },
#            lookupdict=self.cntrlDict
#            ),
#       'geometry_optimization_geometry_change' : MetaInfoMap(startpage),
        'geometry_optimization_method' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['integrator', '== \"steep\"']],
                 'assign' : 'SD'},
                {'test' : [['integrator', '== \"cg\"']],
                 'assign' : 'CG'},
                {'test' : [['integrator', '== \"l-bfgs\"']],
                 'assign' : 'L-BFGS'},
                ],
            lookupdict=self.cntrlDict,
            activeInfo=True,
            ),
#       'geometry_optimization_threshold_force' : MetaInfoMap(startpage),
        'x_gromacs_barostat_type' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['integrator', ' in [\"md\",'
                                          '\"sd\",'
                                          '\"sd2\",'
                                          '\"bd\",'
                                          '\"md-vv\",'
                                          '\"md-vv-avek\"]'],
                           ['pcoupl', '== \"berendsen\"']],
                 'assign' : 'Berendsen'},
                {'test' : [['integrator', ' in [\"md\",'
                                          '\"sd\",'
                                          '\"sd2\",'
                                          '\"bd\",'
                                          '\"md-vv\",'
                                          '\"md-vv-avek\"]'],
                           ['pcoupl', '== \"Parinello-Rahman\"']],
                 'assign' : 'Parinello-Rahman'}
                ],
            lookupdict=self.cntrlDict,
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_barostat']
            ),
        'x_gromacs_barostat_target_pressure' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['integrator', ' in [\"md\",'
                                          '\"sd\",'
                                          '\"sd2\",'
                                          '\"bd\",'
                                          '\"md-vv\",'
                                          '\"md-vv-avek\"]'],
                           ['pcoupl', 'not in [\"no\",\"No\",\"NO\"]']],
                 'value' : 'ref-p\[    0\]'}
                ],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromacs_inout_control_refp0'
                    )
                ),
            lookupdict=self.cntrlDict,
            valtype='float',
            unitdict=self.unitDict,
            unit='bar',
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_barostat']
            ),
        'x_gromacs_barostat_tau' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['integrator', ' in [\"md\",'
                                          '\"sd\",'
                                          '\"sd2\",'
                                          '\"bd\",'
                                          '\"md-vv\",'
                                          '\"md-vv-avek\"]'],
                           ['pcoupl', ' not in [\"no\",\"No\",\"NO\"]']],
                 'value' : 'tau-p'}
                ],
            lookupdict=self.cntrlDict,
            valtype='float',
            unitdict=self.unitDict,
            unit='pico-second',
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_barostat']
            ),
        'x_gromacs_integrator_type' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['integrator', ' in [\"md\",'
                                          '\"sd\",'
                                          '\"sd2\",'
                                          '\"md-vv\",'
                                          '\"md-vv-avek\"]']],
                 'assign' : 'verlet'},
                {'test' : [['integrator', ' in [\"steep\",'
                                          '\"cg\",'
                                          '\"l-bfgs\"]']],
                 'assign' : 'minimization'},
                {'test' : [['integrator', '== \"bd\"']],
                 'assign' : 'Langevin'}
                ],
            lookupdict=self.cntrlDict,
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_integrator']
            ),
        'x_gromacs_integrator_dt' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['integrator', ' in [\"md\",'
                                          '\"sd\",'
                                          '\"sd2\",'
                                          '\"bd\",'
                                          '\"md-vv\",'
                                          '\"md-vv-avek\"]']],
                 'value' : 'dt'}
                ],
            lookupdict=self.cntrlDict,
            valtype='float',
            unitdict=self.unitDict,
            unit='pico-second',
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_integrator']
            ),
        'x_gromacs_number_of_steps_requested' : MetaInfoMap(startpage,
            depends=[{'value' : 'nsteps'}],
            lookupdict=self.cntrlDict,
            valtype='int',
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_integrator']
            ),
        'x_gromacs_thermostat_type' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['integrator', ' in [\"md\",'
                                          '\"sd\",'
                                          '\"sd2\",'
                                          '\"bd\",'
                                          '\"md-vv\",'
                                          '\"md-vv-avek\"]'],
                           ['tcouple', '== \"berendsen\"']],
                 'assign' : 'Berendsen'},
                {'test' : [['integrator', ' in [\"md\",'
                                          '\"sd\",'
                                          '\"sd2\",'
                                          '\"bd\",'
                                          '\"md-vv\",'
                                          '\"md-vv-avek\"]'],
                           ['tcouple', '== \"nose-hoover\"']],
                 'assign' : 'Nose-Hoover'},
                {'test' : [['integrator', ' in [\"md\",'
                                          '\"sd\",'
                                          '\"sd2\",'
                                          '\"bd\",'
                                          '\"md-vv\",'
                                          '\"md-vv-avek\"]'],
                           ['tcouple', '== \"andersen\"']],
                 'assign' : 'Andersen'},
                {'test' : [['integrator', ' in [\"md\",'
                                          '\"sd\",'
                                          '\"sd2\",'
                                          '\"bd\",'
                                          '\"md-vv\",'
                                          '\"md-vv-avek\"]'],
                           ['tcouple', '== \"anderson-massive\"']],
                 'assign' : 'Andersen-Massive'},
                {'test' : [['integrator', ' in [\"md\",'
                                          '\"sd\",'
                                          '\"sd2\",'
                                          '\"bd\",'
                                          '\"md-vv\",'
                                          '\"md-vv-avek\"]'],
                           ['tcouple', '== \"v-rescale\"']],
                 'assign' : 'Velocity Rescaling'},
                ],
            lookupdict=self.cntrlDict,
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_thermostat']
            ),
        'x_gromacs_thermostat_target_temperature' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['integrator', ' in [\"md\",'
                                          '\"sd\",'
                                          '\"sd2\",'
                                          '\"bd\",'
                                          '\"md-vv\",'
                                          '\"md-vv-avek\"]'],
                           ['tcoupl', ' not in [\"no\",\"No\",\"NO\"]']],
                 'value' : 'ref-t'}
                ],
            lookupdict=self.cntrlDict,
            valtype='float',
            unitdict=self.unitDict,
            unit='kelvin',
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_thermostat']
            ),
        'x_gromacs_thermostat_tau' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['integrator', ' in [\"md\",'
                                          '\"sd\",'
                                          '\"sd2\",'
                                          '\"bd\",'
                                          '\"md-vv\",'
                                          '\"md-vv-avek\"]'],
                           ['tcoupl', ' not in [\"no\",\"No\",\"NO\"]']],
                 'value' : 'tau-t'},
                ],
            lookupdict=self.cntrlDict,
            valtype='float',
            unitdict=self.unitDict,
            unit='pico-second',
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_thermostat']
            ),
        'x_gromacs_langevin_bd_fric' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['integrator', '== \"bd\"']],
                 'value' : 'bd-fric'}
                ],
            lookupdict=self.cntrlDict,
            valtype='float',
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_thermostat']
            ),
        'x_gromacs_periodicity_type' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['pbc', ' in [\"no\",\"No\",\"NO\"]']],
                 'assign' : 'no periodic boundaries'},
                {'test' : [['pbc', ' not in [\"no\",\"No\",\"NO\"]']],
                 'assign' : 'periodic boundaries'}
                ],
            lookupdict=self.cntrlDict,
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_thermostat']
            )
        }

    # ------------------------------------------------------------
    #   Definitions for section_single_configuration_calculation
    # ------------------------------------------------------------
    singleconfcalc = {
        #'atom_forces_type' : MetaInfoMap(startpage,
        #    depends=[{'assign' : 'Force Field'}],
        #    lookupdict=self.mddataDict
        #    ),
        'energy_correction_entropy' : MetaInfoMap(startpage),
        #'energy_current' : MetaInfoMap(startpage,
        #    depends=[{'value' : 'EPtot'}],
        #    lookupdict=self.mddataDict
        #    ),
        'energy_current' : MetaInfoMap(startpage),
        'energy_electrostatic' : MetaInfoMap(startpage,
            depends=[{'value' : 'Potential'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            activeInfo=True,
            lookupdict=self.mddataDict
            ),
        'energy_free_per_atom' : MetaInfoMap(startpage),
        'energy_free' : MetaInfoMap(startpage),
        #'energy_method_current' : MetaInfoMap(startpage,
        #    depends=[{'assign' : 'Force Field'}],
        #    lookupdict=self.mddataDict
        #    ),
        'energy_T0_per_atom' : MetaInfoMap(startpage),
        'energy_total_T0_per_atom' : MetaInfoMap(startpage),
        'energy_total_T0' : MetaInfoMap(startpage),
        'energy_total' : MetaInfoMap(startpage,
            depends=[{'value' : 'Total Energy'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            activeInfo=True,
            lookupdict=self.mddataDict
            ),
        'hessian_matrix' : MetaInfoMap(startpage),
        'single_configuration_calculation_converged' : MetaInfoMap(startpage),
        'single_configuration_calculation_to_system_ref' : MetaInfoMap(startpage),
        'single_configuration_to_calculation_method_ref' : MetaInfoMap(startpage),
        'time_calculation' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_cpu1_end' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_cpu1_start' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_date_end' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_date_start' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_wall_end' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_wall_start' : MetaInfoMap(startpage),
#        'stress_tensor_kind' : MetaInfoMap(startpage),
#        'stress_tensor_kind' : MetaInfoMap(startpage,
#            depends=[
#                {'test' : [['integrator', '== \"steep\"']],
#                 'assign' : 'geometry_optimization'},
#                {'test' : [['integrator', '== \"cg\"']],
#                 'assign' : 'geometry_optimization'},
#                {'test' : [['integrator', '== \"l-bfgs\"']],
#                 'assign' : 'geometry_optimization'},
#                {'test' : [['integrator', '== \"md\"']],
#                 'assign' : 'molecular_dynamics'},
#                {'test' : [['integrator', '== \"md-vv\"']],
#                 'assign' : 'molecular_dynamics'},
#                {'test' : [['integrator', '== \"md-vv-avek\"']],
#                 'assign' : 'molecular_dynamics'}
#                ],
#            activeInfo=True,
#            lookupdict=self.cntrlDict
#            ),
        'stress_tensor_value' : MetaInfoMap(startpage)
        }

    # section_single_energy_van_der_Waals of section_single_configuration_calculation
    singlevdw = {
        'energy_van_der_Waals_value' : MetaInfoMap(startpage,
            depends=[{'value' : 'VDWAALS'}],
            lookupdict=self.mddataDict,
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            #autoSections=True,
            activeSections=['section_energy_van_der_Waals']
            ),
        'x_gromacs_energy_lj_14_value' : MetaInfoMap(startpage,
            depends=[{'value' : 'LJ-14'}],
            lookupdict=self.mddataDict,
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            #autoSections=True,
            activeSections=['section_energy_van_der_Waals']
            ),
        }

    # ------------------------------------------
    #   Definitions for section_frame_sequence
    # ------------------------------------------
    frameseq = {
        'frame_sequence_conserved_quantity_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_conserved_quantity_stats' : MetaInfoMap(startpage),
        'frame_sequence_conserved_quantity' : MetaInfoMap(startpage,
            depends=[{'store' : 'Conserved En.'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_continuation_kind' : MetaInfoMap(startpage),
        'frame_sequence_external_url' : MetaInfoMap(startpage),
        'frame_sequence_kinetic_energy_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_kinetic_energy_stats' : MetaInfoMap(startpage),
        'frame_sequence_kinetic_energy' : MetaInfoMap(startpage,
            depends=[{'store' : 'Kinetic En.'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_local_frames_ref' : MetaInfoMap(startpage),
        'frame_sequence_potential_energy_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_potential_energy_stats' : MetaInfoMap(startpage),
        'frame_sequence_potential_energy' : MetaInfoMap(startpage,
            depends=[{'store' : 'Potential'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_pressure_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_pressure_stats' : MetaInfoMap(startpage),
        'frame_sequence_pressure' : MetaInfoMap(startpage,
            depends=[{'store' : 'Pressure (bar)'}],
            valtype='float',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_temperature_frames' : MetaInfoMap(startpage,
            depends=[{'test'  : [['Volume', ' is not None']],
                      'store' : 'Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_temperature_stats' : MetaInfoMap(startpage),
        'frame_sequence_temperature' : MetaInfoMap(startpage,
            depends=[{'store' : 'Temperature'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='Kelvin',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_time' : MetaInfoMap(startpage,
            depends=[{'store' : 'Time'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='pico-second',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_volume_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'test'  : [['Volume', ' is not None']],
                      'store' : 'Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_volume' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Volume'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='Angstrom**3',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_density_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_density' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Density'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='1.0/Angstrom**3',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_bond_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_bond_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Bond'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_ubond_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_ubond_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'U-B'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_proper_dihedral_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_proper_dihedral_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Proper Dih.'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_improper_dihedral_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_improper_dihedral_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Improper Dih.'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_cmap_dihedral_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_cmap_dihedral_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'CMAP Dih.'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_lj_14_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_lj_14_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'LJ-14'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_coulomb_14_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_coulomb_14_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Coulomb-14'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_lj_sr_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_lj_sr_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'LJ (SR)'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_coulomb_sr_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_coulomb_sr_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Coulomb (SR)'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_constrain_rmsd_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromacs_frame_sequence_constrain_rmsd' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Constr. rmsd'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        #'frame_sequence_to_sampling_ref' : MetaInfoMap(startpage),
        'geometry_optimization_converged' : MetaInfoMap(startpage,
            value=self.minConverged
            ),
        #'previous_sequence_ref' : MetaInfoMap(startpage)
        }

    frameseqend = {
        'number_of_conserved_quantity_evaluations_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'frame_sequence_conserved_quantity_frames'
                    )
                )
            ),
        'number_of_frames_in_sequence' : MetaInfoMap(startpage, activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'frame_sequence_potential_energy_frames'
                    )
                )
            ),
        'number_of_kinetic_energies_in_sequence' : MetaInfoMap(startpage, activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'frame_sequence_kinetic_energy_frames'
                    )
                )
            ),
        'number_of_potential_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'frame_sequence_potential_energy_frames'
                    )
                )
            ),
        'number_of_pressure_evaluations_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'frame_sequence_pressure_frames'
                    )
                )
            ),
        'number_of_temperatures_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'frame_sequence_temperature_frames'
                    )
                )
            ),
        'x_gromacs_number_of_volumes_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromacs_frame_sequence_volume_frames'
                    )
                )
            ),
        'x_gromacs_number_of_densities_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromacs_frame_sequence_density_frames'
                    )
                )
            ),
        'x_gromacs_number_of_bond_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromacs_frame_sequence_bond_energy_frames'
                    )
                )
            ),
        'x_gromacs_number_of_ubond_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromacs_frame_sequence_ubond_energy_frames'
                    )
                )
            ),
        'x_gromacs_number_of_coulomb_sr_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromacs_frame_sequence_coulomb_sr_energy_frames'
                    )
                )
            ),
        'x_gromacs_number_of_coulomb_14_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromacs_frame_sequence_coulomb_14_energy_frames'
                    )
                )
            ),
        'x_gromacs_number_of_lj_14_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromacs_frame_sequence_lj_14_energy_frames'
                    )
                )
            ),
        'x_gromacs_number_of_lj_sr_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromacs_frame_sequence_lj_sr_energy_frames'
                    )
                )
            ),
        'x_gromacs_number_of_proper_dihedral_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromacs_frame_sequence_proper_dihedral_energy_frames'
                    )
                )
            ),
        'x_gromacs_number_of_improper_dihedral_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromacs_frame_sequence_improper_dihedral_energy_frames'
                    )
                )
            ),
        'x_gromacs_number_of_cmap_dihedral_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromacs_frame_sequence_cmap_dihedral_energy_frames'
                    )
                )
            ),
        'x_gromacs_number_of_constrain_rmsd_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromacs_frame_sequence_constrain_rmsd_frames'
                    )
                )
            ),
        #'previous_sequence_ref' : MetaInfoMap(startpage)
        }

    # ----------------------------------
    #   Definitions for section_system
    # ----------------------------------
    # section_system
    sec_system = {
        #'topology_ref' : MetaInfoMap(startpage),
        'atom_velocities' : MetaInfoMap(startpage),
        'configuration_raw_gid' : MetaInfoMap(startpage),
        'local_rotations' : MetaInfoMap(startpage),
        'number_of_atoms' : MetaInfoMap(startpage,
            depends=[{'value' : 'number_of_atoms'}],
            valtype='int',
            lookupdict=self.extraDict
            ),
        'number_of_sites' : MetaInfoMap(startpage),
        'number_of_symmetry_operations' : MetaInfoMap(startpage),
        'reduced_symmetry_matrices' : MetaInfoMap(startpage),
        'reduced_symmetry_translations' : MetaInfoMap(startpage),
        'SC_matrix' : MetaInfoMap(startpage),
        'spacegroup_3D_choice' : MetaInfoMap(startpage),
        'spacegroup_3D_hall' : MetaInfoMap(startpage),
        'spacegroup_3D_international' : MetaInfoMap(startpage),
        'spacegroup_3D_number' : MetaInfoMap(startpage),
        'spacegroup_3D_origin_shift' : MetaInfoMap(startpage),
        'spacegroup_3D_pointgroup' : MetaInfoMap(startpage),
        'spacegroup_3D_std_lattice' : MetaInfoMap(startpage),
        'spacegroup_3D_std_positions' : MetaInfoMap(startpage),
        'spacegroup_3D_std_types' : MetaInfoMap(startpage),
        'spacegroup_3D_trasformation_matrix' : MetaInfoMap(startpage),
        'spacegroup_3D_wyckoff' : MetaInfoMap(startpage),
        'symmorphic' : MetaInfoMap(startpage),
        'system_name' : MetaInfoMap(startpage,
            subfunction={
                'function' : MDData.MDDataConverter.topology_system_name,
                'supportDict' : self.topology.topoDict,
                },
            #functionbase=self
            ),
        'time_reversal_symmetry' : MetaInfoMap(startpage)
        }

    # section_configuration_core of section_system
    configuration_core = {
        'number_of_electrons' : MetaInfoMap(startpage,
            value=0,
            valtype='float',
            ),
        'atom_labels' : MetaInfoMap(startpage,
            #subfunction=self.system_atom_labels()
            ),
        'atom_positions' : MetaInfoMap(startpage,
            #subfunction=self.system_atom_positions()
            ),
        'configuration_periodic_dimensions' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['pbc', '== "no"']],
                 'assign' : np.asarray([False, False, False])},
                {'test' : [['pbc', '== "xy"']],
                 'assign' : np.asarray([True, True, False])},
                {'test' : [['pbc', '== "xyz"']],
                 'assign' : np.asarray([True, True, True])}
                ],
            lookupdict=self.cntrlDict,
            ),
        'embedded_system' : MetaInfoMap(startpage),
        'lattice_vectors' : MetaInfoMap(startpage,
            #subfunction=self.system_lattice_vectors()
            ),
        'simulation_cell' : MetaInfoMap(startpage,
            #subfunction=self.system_simulation_cell()
            )
        }

    # section_spacegroup_3D_operation of section_system
    spacegroup_op = {
            'spacegroup_3D_rotation' : MetaInfoMap(startpage),
            'spacegroup_3D_translation' : MetaInfoMap(startpage)
        }

    # section_system_to_system_refs of section_system
    sys_to_sys = {
            'system_to_system_kind' : MetaInfoMap(startpage),
            'system_to_system_ref' : MetaInfoMap(startpage)
        }

    # --------------------------------------------------------
    #   Definitions of meta data values for section_topology
    # --------------------------------------------------------
    # section_topology of section_run
    topology = {
        'atom_to_molecule' : MetaInfoMap(startpage,
            subfunction={
                'function' : MDData.MDDataConverter.topology_atom_to_mol,
                'supportDict' : self.topology.topoDict,
                },
            #functionbase=self
            ),
        'molecule_to_molecule_type_map' : MetaInfoMap(startpage),
        'number_of_topology_atoms' : MetaInfoMap(startpage,
            depends=[{'value' : 'NATOM'}],
            valtype='int',
            lookupdict=self.parmDict
            ),
        'number_of_topology_molecules' : MetaInfoMap(startpage,
            subfunction={
                'function' : MDData.MDDataConverter.topology_num_topo_mol,
                'supportDict' : self.topology.topoDict,
                },
            valtype='int',
            ),
        'topology_force_field_name' : MetaInfoMap(startpage,
            value='Force Field',
            )
        }

    # section_atom_type of section_topology
    atom_type = {
        'atom_type_charge' : MetaInfoMap(startpage),
        'atom_type_mass' : MetaInfoMap(startpage),
        'atom_type_name' : MetaInfoMap(startpage)
        }

    # section_constraint of section_topology
    constraint = {
        'constraint_atoms' : MetaInfoMap(startpage),
        'constraint_kind' : MetaInfoMap(startpage),
        'constraint_parameters' : MetaInfoMap(startpage),
        'number_of_atoms_per_constraint' : MetaInfoMap(startpage),
        'number_of_constraints' : MetaInfoMap(startpage)
        }

    # section_interaction of section_topology
    interaction = {
        'interaction_atoms' : MetaInfoMap(startpage),
        'interaction_kind' : MetaInfoMap(startpage),
        'interaction_parameters' : MetaInfoMap(startpage),
        'number_of_atoms_per_interaction' : MetaInfoMap(startpage),
        'number_of_interactions' : MetaInfoMap(startpage)
        }

    # -------------------------------------------------------------
    #   Definitions of meta data values for section_molecule_type
    # -------------------------------------------------------------
    # section_molecule_type of section_topology
    mol_type = {
        'molecule_type_name' : MetaInfoMap(startpage),
        'number_of_atoms_in_molecule' : MetaInfoMap(startpage),
        'settings_atom_in_molecule' : MetaInfoMap(startpage)
        }

    # section_molecule_constraint of section_molecule_type
    mol_constraint = {
        'molecule_constraint_atoms' : MetaInfoMap(startpage),
        'molecule_constraint_kind' : MetaInfoMap(startpage),
        'molecule_constraint_parameters' : MetaInfoMap(startpage),
        'number_of_atoms_per_molecule_constraint' : MetaInfoMap(startpage),
        'number_of_molecule_constraints' : MetaInfoMap(startpage)
        }

    # section_molecule_interaction of section_molecule_type
    mol_interaction = {
        'molecule_interaction_atoms' : MetaInfoMap(startpage),
        'molecule_interaction_kind' : MetaInfoMap(startpage),
        'molecule_interaction_parameters' : MetaInfoMap(startpage),
        'number_of_atoms_per_molecule_interaction' : MetaInfoMap(startpage),
        'number_of_molecule_interactions' : MetaInfoMap(startpage)
        }

    # section_atom_in_molecule of section_molecule_type
    atom_in_mol = {
        'atom_in_molecule_charge' : MetaInfoMap(startpage),
        'atom_in_molecule_name' : MetaInfoMap(startpage),
        'atom_in_molecule_to_atom_type_ref' : MetaInfoMap(startpage)
        }


    if defname == 'system':
        dictionary = systemDict
    elif defname == 'topology':
        dictionary = topology
    elif defname == 'singleconfcalc':
        dictionary = singleconfcalc
    elif defname == 'frameseq':
        dictionary = frameseq
    elif defname == 'frameseqend':
        dictionary = frameseqend
    elif defname == 'atom_type':
        dictionary = atom_type
    elif defname == 'molecule_type':
        dictionary = moltypeDict
    elif defname == 'interaction':
        dictionary = interaction
    elif defname == 'sampling':
        dictionary = sampling
    elif defname == 'singlevdw':
        dictionary = singlevdw
    else:
        dictionary = singleconfcalclist
    return MapDictionary(dictionary)




# pylint: disable=relative-beyond-top-level
import unittest

from optimade.models import ReferenceResponseMany, ReferenceResponseOne

from ..utils import EndpointTestsMixin


class ReferencesEndpointTests(EndpointTestsMixin, unittest.TestCase):

    request_str = "/references"
    response_cls = ReferenceResponseMany


class SingleReferenceEndpointTests(EndpointTestsMixin, unittest.TestCase):

    test_id = "dijkstra1968"
    request_str = f"/references/{test_id}"
    response_cls = ReferenceResponseOne


class SingleReferenceEndpointTestsDifficult(EndpointTestsMixin, unittest.TestCase):

    test_id = "dummy/20.19"
    request_str = f"/references/{test_id}"
    response_cls = ReferenceResponseOne



# pylint: disable=line-too-long,no-self-argument
from pydantic import Field, BaseModel, AnyUrl, validator
from typing import List, Optional

from .entries import EntryResource, EntryResourceAttributes


__all__ = ("Person", "ReferenceResourceAttributes", "ReferenceResource")


class Person(BaseModel):
    name: str = Field(..., decsription="""Full name of the person, REQUIRED.""")
    firstname: Optional[str] = Field(None, description="""First name of the person.""")
    lastname: Optional[str] = Field(None, description="""Last name of the person.""")


class ReferenceResourceAttributes(EntryResourceAttributes):
    """ Model that stores the attributes of a reference. Many properties match the
    meaning described in the
    [BibTeX specification](http://bibtexml.sourceforge.net/btxdoc.pdf).

    """

    authors: Optional[List[Person]] = Field(
        None,
        description="List of person objects containing the authors of the reference.",
    )
    editors: Optional[List[Person]] = Field(
        None,
        description="List of person objects containing the editors of the reference.",
    )

    doi: Optional[str] = Field(
        None, description="The digital object identifier of the reference."
    )

    url: Optional[AnyUrl] = Field(None, description="The URL of the reference.")

    address: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    annote: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    booktitle: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    chapter: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    crossref: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    edition: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    howpublished: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    institution: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    journal: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    key: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    month: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    note: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    number: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    organization: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    pages: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    publisher: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    school: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    series: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    title: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    bib_type: Optional[str] = Field(
        None,
        description="Type of the reference, corresponding to the **type** property in the BiBTeX specification.",
    )
    volume: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )
    year: Optional[str] = Field(
        None, description="Meaning of property matches the BiBTeX specification."
    )


class ReferenceResource(EntryResource):
    """ The :entry:`references` entries describe bibliographic references.
The following properties are used to provide the bibliographic details:

- **address**, **annote**, **booktitle**, **chapter**, **crossref**, **edition**, **howpublished**, **institution**, **journal**, **key**, **month**,
  **note**, **number**, **organization**, **pages**, **publisher**, **school**, **series**, **title**, **type**, **volume**, **year**:
  Meanings of these properties match the `BibTeX specification <http://bibtexml.sourceforge.net/btxdoc.pdf>`__, values are strings;

- **authors** and **editors**: lists of *person objects* which are dictionaries with the following keys:

  - **name**: Full name of the person, REQUIRED.
  - **firstname**, **lastname**: Parts of the person's name, OPTIONAL.

- **doi** and **url**: values are strings.

- **Requirements/Conventions**:

  - **Support**: OPTIONAL, i.e., any of the properties MAY be :val:`null`.
  - **Query**: Support for queries on any of these properties is OPTIONAL.
    If supported, filters MAY support only a subset of comparison operators.
  - Every references entry MUST contain at least one of the properties."""

    type: str = Field(
        default="references",
        const=True,
        description="""The name of the type of an entry. Any entry MUST be able to be fetched using the `base URL <Base URL_>`_ type and ID at the url :endpoint:`<base URL>/<type>/<id>`.
- **Type**: string.
- **Requirements/Conventions**:

  - **Support**: REQUIRED, MUST NOT be :val:`null`.
  - **Query**: MUST be a queryable property with support for all mandatory filter features.
  - **Response**: REQUIRED in the response.
  - MUST be an existing entry type.
  - The entry of type `<type>` and ID `<id>` MUST be returned in response to a request for :endpoint:`/<type>/<id>` under the versioned base URL.

- **Example**: :val:`"structures"`""",
    )
    attributes: ReferenceResourceAttributes

    @validator("attributes")
    def validate_attributes(cls, v):
        if not any(prop[1] is not None for prop in v):
            raise ValueError("reference object must have at least one field defined")
        return v



# Copyright 2018 Markus Scheidgen, empty_task
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This module describes all configurable parameters for the nomad python code. The
configuration is used for all executed python code including API, worker, CLI, and other
scripts. To use the configuration in your own scripts or new modules, simply import
this module.

All parameters are structured into objects for two reasons. First, to have
categories. Second, to allow runtime manipulation that is not effected
by python import logic. The categories are choosen along infrastructure components:
``mongo``, ``elastic``, etc.

This module also provides utilities to read the configuration from environment variables
and .yaml files. This is done automatically on import. The precedence is env over .yaml
over defaults.

.. autoclass:: nomad.config.NomadConfig
.. autofunction:: nomad.config.apply
.. autofunction:: nomad.config.load_config
'''

import logging
import os
import os.path
import yaml
import warnings

try:
    from nomad import gitinfo
except ImportError:
    git_root = os.path.join(os.path.dirname(__file__), '..')
    cwd = os.getcwd()
    os.chdir(git_root)
    os.system('./gitinfo.sh')
    os.chdir(cwd)

    from nomad import gitinfo


warnings.filterwarnings('ignore', message='numpy.dtype size changed')
warnings.filterwarnings('ignore', message='numpy.ufunc size changed')


class NomadConfig(dict):
    '''
    A class for configuration categories. It is a dict subclass that uses attributes as
    key/value pairs.
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError("No such attribute: " + name)


CELERY_WORKER_ROUTING = 'worker'
CELERY_QUEUE_ROUTING = 'queue'

rabbitmq = NomadConfig(
    host='localhost',
    user='rabbitmq',
    password='rabbitmq'
)


def rabbitmq_url():
    return 'pyamqp://%s:%s@%s//' % (rabbitmq.user, rabbitmq.password, rabbitmq.host)


celery = NomadConfig(
    max_memory=64e6,  # 64 GB
    timeout=1800,  # 1/2 h
    acks_late=True,
    routing=CELERY_QUEUE_ROUTING,
    priorities={
        'Upload.process_upload': 5,
        'Upload.delete_upload': 9,
        'Upload.publish_upload': 10
    }
)

fs = NomadConfig(
    tmp='.volumes/fs/tmp',
    staging='.volumes/fs/staging',
    public='.volumes/fs/public',
    local_tmp='/tmp',
    prefix_size=2,
    working_directory=os.getcwd()
)

elastic = NomadConfig(
    host='localhost',
    port=9200,
    index_name='nomad_fairdi_calcs'
)

keycloak = NomadConfig(
    server_url='https://nomad-lab.eu/fairdi/keycloak/auth/',
    realm_name='fairdi_nomad_test',
    username='admin',
    password='password',
    client_id='nomad_public',
    client_secret=None,
    oasis=False)

mongo = NomadConfig(
    host='localhost',
    port=27017,
    db_name='nomad_fairdi'
)

logstash = NomadConfig(
    enabled=False,
    host='localhost',
    tcp_port='5000',
    level=logging.DEBUG
)

services = NomadConfig(
    api_host='localhost',
    api_port=8000,
    api_base_path='/fairdi/nomad/latest',
    api_secret='defaultApiSecret',
    api_chaos=0,
    admin_user_id='00000000-0000-0000-0000-000000000000',
    not_processed_value='not processed',
    unavailable_value='unavailable',
    https=False,
    upload_limit=10,
    force_raw_file_decoding=False,
    download_scan_size=500,
    download_scan_timeout=u'30m'
)

tests = NomadConfig(
    default_timeout=30
)


def api_url(ssl: bool = True):
    return '%s://%s/%s/api' % (
        'https' if services.https and ssl else 'http',
        services.api_host.strip('/'),
        services.api_base_path.strip('/'))


def gui_url(page: str = None):
    base = api_url(True)[:-3]
    if base.endswith('/'):
        base = base[:-1]

    if page is not None:
        return '%s/gui/%s' % (base, page)

    return '%s/gui' % base


def check_config():
    """Used to check that the current configuration is valid. Should only be
    called once after the final config is loaded.

    Raises:
        AssertionError: if there is a contradiction or invalid values in the
            config file settings.
    """
    # The AFLOW symmetry information is checked once on import
    proto_symmetry_tolerance = normalize.prototype_symmetry_tolerance
    symmetry_tolerance = normalize.symmetry_tolerance
    if proto_symmetry_tolerance != symmetry_tolerance:
        raise AssertionError(
            "The AFLOW prototype information is outdated due to changed tolerance "
            "for symmetry detection. Please update the AFLOW prototype information "
            "by running the CLI command 'nomad admin ops prototype-update "
            "--matches-only'."
        )

    if not os.path.exists(normalize.springer_db_path):
        normalize.springer_db_path = None


mail = NomadConfig(
    enabled=False,
    with_login=False,
    host='',
    port=8995,
    user='',
    password='',
    from_address='support@nomad-lab.eu',
    cc_address='support@nomad-lab.eu'
)

normalize = NomadConfig(
    # The system size limit for running the dimensionality analysis. For very
    # large systems the dimensionality analysis will get too expensive.
    system_classification_with_clusters_threshold=64,
    # Symmetry tolerance controls the precision used by spglib in order to find
    # symmetries. The atoms are allowed to move 1/2*symmetry_tolerance from
    # their symmetry positions in order for spglib to still detect symmetries.
    # The unit is angstroms. The value of 0.1 is used e.g. by Materials Project
    # according to
    # https://pymatgen.org/pymatgen.symmetry.analyzer.html#pymatgen.symmetry.analyzer.SpacegroupAnalyzer
    symmetry_tolerance=0.1,
    # The symmetry tolerance used in aflow prototype matching. Should only be
    # changed before re-running the prototype detection.
    prototype_symmetry_tolerance=0.1,
    # Maximum number of atoms in the single cell of a 2D material for it to be
    # considered valid.
    max_2d_single_cell_size=7,
    # The distance tolerance between atoms for grouping them into the same
    # cluster. Used in detecting system type.
    cluster_threshold=2.5,
    # Defines the "bin size" for rounding cell angles for the material hash
    angle_rounding=float(10.0),  # unit: degree
    # The threshold for a system to be considered "flat". Used e.g. when
    # determining if a 2D structure is purely 2-dimensional to allow extra rigid
    # transformations that are improper in 3D but proper in 2D.
    flat_dim_threshold=0.1,
    # The threshold for point equality in k-space. Unit: 1/m.
    k_space_precision=150e6,
    # The energy threshold for how much a band can be on top or below the fermi
    # level in order to detect a gap. Unit: Joule.
    band_structure_energy_tolerance=1.6022e-20,  # 0.1 eV
    springer_db_path=os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'normalizing/data/springer.msg'
    )
)

paths = NomadConfig(
    similarity="",
)

client = NomadConfig(
    user='leonard.hofstadter@nomad-fairdi.tests.de',
    password='password',
    url='http://nomad-lab.eu/prod/rae/api'
)

datacite = NomadConfig(
    mds_host='https://mds.datacite.org',
    enabled=False,
    prefix='10.17172',
    user='*',
    password='*'
)

meta = NomadConfig(
    version='0.9.1',
    commit=gitinfo.commit,
    release='devel',
    default_domain='dft',
    service='unknown nomad service',
    name='novel materials discovery (NOMAD)',
    description='A FAIR data sharing platform for materials science data',
    homepage='https://nomad-lab.eu',
    source_url='https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-FAIR',
    maintainer_email='markus.scheidgen@physik.hu-berlin.de'
)

auxfile_cutoff = 100
parser_matching_size = 9128
console_log_level = logging.WARNING
max_upload_size = 32 * (1024 ** 3)
raw_file_strip_cutoff = 1000
use_empty_parsers = False
reprocess_unmatched = True
metadata_file_name = 'nomad'
metadata_file_extensions = ('json', 'yaml', 'yml')


def normalize_loglevel(value, default_level=logging.INFO):
    plain_value = value
    if plain_value is None:
        return default_level
    else:
        try:
            return int(plain_value)
        except ValueError:
            return getattr(logging, plain_value)


transformations = {
    'console_log_level': normalize_loglevel,
    'logstash_level': normalize_loglevel
}


# use std python logger, since logging is not configured while loading configuration
logger = logging.getLogger(__name__)


def apply(key, value) -> None:
    '''
    Changes the config according to given key and value. The keys are interpreted as paths
    to config values with ``_`` as a separator. E.g. ``fs_staging`` leading to
    ``config.fs.staging``
    '''
    path = list(reversed(key.split('_')))
    child_segment = None
    current_value = None
    child_config = globals()
    child_key = None

    try:
        while len(path) > 0:
            if child_segment is None:
                child_segment = path.pop()
            else:
                child_segment += '_' + path.pop()

            if child_segment in child_config:
                current_value = child_config[child_segment]

            if current_value is None:
                if len(path) == 0:
                    raise KeyError

                continue
            if isinstance(current_value, NomadConfig):
                child_config = current_value
                current_value = None
                child_segment = None
            else:
                if len(path) > 0:
                    raise KeyError()

                child_key = child_segment
                break

        if child_key is None or current_value is None:
            raise KeyError()
    except KeyError:
        return

    if not isinstance(value, type(current_value)):
        try:
            value = transformations.get(key, type(current_value))(value)
        except Exception as e:
            logger.error(
                'config key %s value %s has wrong type: %s' % (key, str(value), str(e)))

    child_config[child_key] = value


def load_config(config_file: str = os.environ.get('NOMAD_CONFIG', 'nomad.yaml')) -> None:
    '''
    Loads the configuration from the ``config_file`` and environment.

    Arguments:
        config_file: Override the configfile, default is file stored in env variable
            NOMAD_CONFIG or ``nomad.yaml``.
    '''
    # load yaml and override defaults (only when not in test)
    if os.path.exists(config_file):
        with open(config_file, 'r') as stream:
            try:
                config_data = yaml.load(stream, Loader=getattr(yaml, 'FullLoader'))
            except yaml.YAMLError as e:
                logger.error('cannot read nomad config', exc_info=e)

        def adapt(config, new_config, child_key=None):
            for key, value in new_config.items():
                if key in config:
                    if child_key is None:
                        qualified_key = key
                    else:
                        qualified_key = '%s_%s' % (child_key, key)

                    current_value = config[key]
                    if isinstance(value, dict) and isinstance(current_value, NomadConfig):
                        adapt(current_value, value, qualified_key)
                    else:
                        if not isinstance(value, type(current_value)):
                            try:
                                value = transformations.get(qualified_key, type(current_value))(value)
                            except Exception as e:
                                logger.error(
                                    'config key %s value %s has wrong type: %s' % (key, str(value), str(e)))
                        else:
                            config[key] = value
                            logger.debug('override config key %s with value %s' % (key, str(value)))
                else:
                    logger.error('config key %s does not exist' % key)

        if config_data is not None:
            adapt(globals(), config_data)

    # load env and override yaml and defaults
    kwargs = {
        key[len('NOMAD_'):].lower(): value
        for key, value in os.environ.items()
        if key.startswith('NOMAD_')
    }

    for key, value in kwargs.items():
        apply(key, value)

    check_config()


load_config()



# pylint: disable=import-error
import pytest

from .utils import get_min_ver

min_ver = get_min_ver("numpy")
numpy = pytest.importorskip(
    "numpy",
    minversion=min_ver,
    reason=f"numpy must be installed with minimum version {min_ver} for these tests to"
    " be able to run",
)

from optimade.adapters import Structure
from optimade.adapters.structures.proteindatabank import get_pdb


def test_successful_conversion(RAW_STRUCTURES):
    """Make sure its possible to convert"""
    for structure in RAW_STRUCTURES:
        assert isinstance(get_pdb(Structure(structure)), str)


def test_null_positions(null_position_structure):
    """Make sure null positions are handled"""
    assert isinstance(get_pdb(null_position_structure), str)


def test_null_lattice_vectors(null_lattice_vector_structure):
    """Make sure null lattice vectors are handled"""
    assert isinstance(get_pdb(null_lattice_vector_structure), str)


def test_special_species(SPECIAL_SPECIES_STRUCTURES):
    """Make sure vacancies and non-chemical symbols ("X") are handled"""
    for special_structure in SPECIAL_SPECIES_STRUCTURES:
        structure = Structure(special_structure)

        assert isinstance(get_pdb(structure), str)



import numpy as np
from contextlib import contextmanager
import logging
import json
import os
import re
import ast
from collections import namedtuple

NOTEXCEPT = re.compile(r'[a-cf-zA-CF-Z!\?,{}\[\]]')

def is_number(val):
        try:
            float(val)
            return True
        except ValueError:
            return False

def strcleaner(val):
        unwantedkeys = [
                "system", "eval", "return",
                "ctypes", "setup", "import",
                "git", "swig", "cython"]
        for keyword in unwantedkeys:
            val = val.replace(keyword, '')
        return val

def strisinstance(val, typ):
        typlist=[]
        if isinstance(typ, str):
            typlist.append(typ)
        elif isinstance(typ, (tuple,list)):
            for ty in typ:
                typlist.append(ty)
        anytype=None
        for t in typlist:
            try:
                if("list" in t and
                   "[" in val and
                   "]" in val and
                   "," in val):
                    if isinstance(literal_eval(strcleaner(val)), list):
                        anytype="list"
                        break
                    elif isinstance(literal_eval(strcleaner(val)), np.ndarray):
                        anytype="np.ndarray"
                        break
                elif("tuple" in t and
                   "(" in val and
                   ")" in val and
                   "," in val):
                    if isinstance(literal_eval(strcleaner(val)), tuple):
                        anytype="tuple"
                        break
            except (TypeError,ValueError,AttributeError):
                pass
        return anytype

def literal_eval(val):
        try:
            return ast.literal_eval(val)
        except (TypeError,ValueError):
            return val

class Container(object):
    """The container class for nested data storage
    """
    def __init__(self, *args):
        self.Name = []
        for arg in args:
            if isinstance(arg, str):
                self.Name.append(arg)
        self.Active = False
        self.OpenBackend = False
        self.CloseBackend = False
        self.gid = None
        self.Containers = []
        self.References = []
        self.ReferencedFrom = []
        self.Storage = None
        self.Indent = '   '
        self.Color = None
        self.PrintOnlyActive = None
        self.PrintOnlyNames = None
        self.localDict = None
        self.currentValue = None

    def add(self, *args):
        for arg in args:
            if isinstance(arg, Container):
                self.Containers.append(arg)
            if isinstance(arg, Storage):
                if self.Storage:
                    self.Storage(arg.__dict__)
                else:
                    self.Storage = arg
            if isinstance(arg, dict):
                if self.Storage:
                    if isinstance(self.Storage, dict):
                        self.Storage.update(arg)
                    elif isinstance(self.Storage, Storage):
                        self.Storage(arg)
                else:
                    self.Storage = Storage(arg)
            if isinstance(arg, JsonMetaInfo):
                for item in arg.jsonList:
                    self.add(item)

    def build(self, metadata, startsection, umaskdict):
        if isinstance(metadata, JsonMetaInfo):
            attrdict = metadata.attributes(startsection)
            if attrdict:
                self.add(attrdict)
            childs = metadata.siblings(startsection)
            if startsection in umaskdict.keys():
                excludes = umaskdict[startsection]
            else:
                excludes = []
            for section in childs:
                if section not in excludes:
                    newContainer = Container(section)
                    newContainer.build(metadata, section, umaskdict)
                    if newContainer.Storage is None:
                        newContainer.add()
                    self.add(newContainer)
            return True
        else:
            return False

    def populate(self, metadata, startsection, umaskdict, updatedict):
        if isinstance(metadata, JsonMetaInfo):
            self.build(metadata, startsection, umaskdict)
            self.update(updatedict)

    def updaterefs(self, *args):
        for arg in args:
            if isinstance(arg, Container):
                self.References.extend(arg.Name)

    def update(self, *args):
        for arg in args:
            if isinstance(arg, dict):
                if arg["startSection"]:
                    if self.Name in arg["startSection"]:
                        self.accumulateValues(self, arg)
                    else:
                        if self.Containers:
                            for module in self.Containers:
                                module.update(arg)
                else:
                    self.accumulateValues(self, arg)

    def reset(self, *args):
        for arg in args:
            if isinstance(arg, dict):
                if arg["startSection"]:
                    if self.Name in arg["startSection"]:
                        self.resetValues(self, arg)
                    else:
                        if self.Containers:
                            for module in self.Containers:
                                module.reset(arg)
                else:
                    self.resetValues(self, arg)

    def updateBackend(self, backend, startsection=None, autoopenclose=False):
        if startsection:
            if self.Name == startsection:
                if autoopenclose:
                    with self.autosection(backend, self.Name):
                        self.updateBackendValues(backend)
                else:
                    self.updateBackendValues(backend)
            else:
                if self.Containers:
                    for module in self.Containers:
                        module.updateBackend(backend, startsection, autoopenclose)
        else:
            if autoopenclose:
                with self.autosection(backend, self.Name):
                    self.updateBackendValues(backend)
            else:
                self.updateBackendValues(backend)

    @contextmanager
    def autosection(self, backend, name):
        self.gid = backend.openSection(name)
        yield self.gid
        backend.closeSection(name, self.gid)

    def opensection(self, backend, name):
        self.gid = backend.openSection(name)
        yield self.gid

    def closesection(self, backend, name):
        backend.closeSection(name, self.gid)

    def fetchAttr(self, resdict):
        for item in resdict:
            if self.Storage:
                if item in self.Storage.__dict__:
                    resdict.update({item: self.Storage.__dict__[item]})
                else:
                    if self.Containers:
                        for module in self.Containers:
                            resdict.update(module.fetchAttr(resdict))
            else:
                if self.Containers:
                    for module in self.Containers:
                        resdict.update(module.fetchAttr(resdict))
        return resdict

    def fetchAttrValue(self, itemName):
        rtnValue=None
        if self.Storage:
            if itemName in self.Storage.__dict__.keys():
                rtnValue=self.Storage.__dict__[itemName]
            else:
                if self.Containers:
                    for module in self.Containers:
                        rtnValue=module.fetchAttrValue(itemName)
                        if rtnValue is not None:
                            break
        else:
            if self.Containers:
                for module in self.Containers:
                    rtnValue=module.fetchAttrValue(itemName)
                    if rtnValue is not None:
                        break
        return rtnValue

    def updateBackendValues(self, backend):
        if self.Storage:
            self.updateBackendStorage(backend)
            self.Active = False
        if self.Containers:
            for module in self.Containers:
                module.updateBackendValues(backend)

    def accumulateValues(self, *args):
        for arg in args:
            if isinstance(arg, dict):
                if self.Storage:
                    self.accumulateDict(arg["dictionary"])
                if self.Containers:
                    for module in self.Containers:
                        module.accumulateValues(arg)
                if "activeSections" in arg:
                    if self.Name in arg["activeSections"]:
                        self.Active = True
                if "muteSections" in arg:
                    if self.Name in arg["muteSections"]:
                        self.Active = False

    def resetValues(self, *args):
        for arg in args:
            if isinstance(arg, dict):
                if self.Storage:
                    self.resetAllValues()
                if self.Containers:
                    for module in self.Containers:
                        module.resetValues(arg)
                self.Active = False

    def checkUpdateValue(self, item, localdict):
        """ Updating value with the rules given in depends

         Updating values follows the following order:
          1) If 'depends' is supplied (not empty or not None),
             the tests in the depends list will be checked in order.
             If one of the tests is successful than the one of the
             values in 'assign' or 'value' will be updated for the item.
             Here 'assign' will assign a new value in the given string and
             'value' will assign the value of the given key item.
             (Ex. : 'assign' : 'CG' will update the item with 'CG' while
             'value' : 'NATOM' will update the item with number of atoms
             returned by the value of NATOM key ,which is stored in lookup dict.)
          2) If 'depends' is supplied but a lookup dictionary is not than
             only the values of attributes in the sections can be used for test.
             The rest of the tests and assignments are updated as in case (1).
          3) If 'depends' is not supplied, subfunction is used to update value.
          4) If 'depends' and subfunction are not supplied but value of
             MetaInfoMap is supplied, the value will be assign directly from the value
             item.
          5) If none of the above items are supplied, this function will return None
             to not update any values for the selected item.
        """
        # Check whether depends is supplied in the item.
        updateValue = None
        storeValue = False
        self.currentValue = updateValue
        if "prefunction" in item:
            prefunc = item.prefunction
            storeValue, updateValue, item = prefunc(item)
            self.currentValue = updateValue
        if "depends" in item:
            firstdepend = item["depends"][0]
            if "lookupdict" in item:
                needFetchVal = False
                if "test" in firstdepend:
                    storeValue, updateValue, localdict = self.checkTestsDicts(item, localdict)
                elif "assign" in firstdepend:
                    updateValue = firstdepend["assign"]
                elif "value" in firstdepend:
                    itemdepval = firstdepend["value"]
                    needFetchVal = True
                elif "store" in firstdepend:
                    itemdepval = firstdepend["store"]
                    needFetchVal = True
                    storeValue = True
                if needFetchVal:
                    if itemdepval in localdict:
                        checkval = localdict[itemdepval]
                    else:
                        accessName, checkval = self.findNameInLookupDict(itemdepval, item.lookupdict)
                        localdict.update({itemdepval : checkval})
                    updateValue = checkval
            else:
                needFetchVal = False
                if "test" in firstdepend:
                    storeValue, updateValue, localdict = self.checkTestsAttr(item, localdict)
                elif "assign" in firstdepend:
                    updateValue = firstdepend["assign"]
                elif "value" in firstdepend:
                    itemdepval = firstdepend["value"]
                    needFetchVal = True
                elif "store" in firstdepend:
                    itemdepval = firstdepend["store"]
                    needFetchVal = True
                    storeValue = True
                if needFetchVal:
                    if itemdepval in localdict:
                        checkval = localdict[itemdepval]
                    else:
                        attrdict = {deptest[0] : ''}
                        attrdict = self.fetchAttr(attrdict)
                        localdict.update(attrdict)
                        checkval = attrdict[deptest[0]]
                    updateValue = checkval
        self.currentValue = updateValue
        if "subfunction" in item:
            subfunc = item.subfunction["function"]
            if "supportDict" in item.subfunction:
                supDict = item.subfunction["supportDict"]
                storeValue, updateValue, item = subfunc(supDict, item)
            else:
                storeValue, updateValue, item = subfunc(item)
            self.currentValue = updateValue
        if "value" in item:
            updateValue = item['value']
            self.currentValue = updateValue
        if("depends" in item or "value" in item or "subfunction" in item):
            pass
        elif "defaultValue" in item:
            updateValue = item['defaultValue']
            self.currentValue = updateValue
        if "postfunction" in item:
            postfunc = item.postfunction
            storeValue, updateValue, item = postfunc(item)
            self.currentValue = updateValue
        if "valtype" in item:
            if updateValue is not None:
                updateValue = self.convertToNumber(updateValue, item["valtype"])
        if("unit" in item and "unitdict" in item):
            if updateValue is not None:
                updateValue = self.convertUnits(updateValue, item["unit"], item["unitdict"])
        return storeValue, updateValue, localdict

    def convertToNumber(self, updateValue, valtype):
        acceptvals = ["float", "int", "list", "tuple",
                      "np.asarray", "np.array",
                      "set", "boolean", "bytes"]
        if valtype in acceptvals:
            if(isinstance(updateValue, list) or isinstance(updateValue, tuple)):
                try:
                    newUpdateValue = [eval(
                        valtype+"("+literal_eval(str(ival))+")"
                        ) for ival in updateValue]
                except (TypeError,ValueError):
                    newUpdateValue = updateValue
            elif isinstance(updateValue, np.ndarray):
                try:
                    newUpdateValue = np.asarray([eval(
                        valtype+"("+literal_eval(str(ival))+")"
                        ) for ival in updateValue])
                except (TypeError,ValueError):
                    newUpdateValue = updateValue
            elif is_number(updateValue):
                newUpdateValue = eval(valtype+"("+str(updateValue)+")")
            else:
                newUpdateValue = updateValue
        else:
            # I hope you know what you are doing
            try:
                newUpdateValue = float(updateValue)
            except (TypeError,ValueError):
                newUpdateValue = updateValue
        return newUpdateValue

    def convertUnits(self, updateValue, unit, unitdict):
        if(isinstance(updateValue, list) or isinstance(updateValue, tuple)):
            updateValue = [float(ival) * self.unitConverter(
                unit, unitdict
                ) for ival in updateValue]
        elif isinstance(updateValue, np.ndarray):
            updateValue = updateValue * self.unitConverter(
                unit, unitdict)
        elif is_number(updateValue):
            updateValue = self.convertToNumber(updateValue, "float")
            if updateValue:
                updateValue = updateValue * self.unitConverter(
                        unit, unitdict)
        elif isinstance(updateValue, str):
            # I hope you know what you are doing
            try:
                newUpdateVal = strcleaner(updateValue)
                newUpdateVal = NOTEXCEPT.sub('', newUpdateVal)
                updateValue = float(newUpdateVal) * self.unitConverter(
                        unit, unitdict)
            except (TypeError,ValueError):
                pass
        return updateValue

    def unitConverter(self, unit, unitdict):
        """ Unit converter using definitions of units explicitly

            The unit names are converted to numbers and the resulting
            expression will be evaluated by python.
            Ex.: unit = 'electron-volt/Angstrom^3'
                 will be converted to
                 unit = '1.602176565e-19*1.0/1.0e-10**3'
                 factor = eval(unit) = 160.2176565e+9 Joule/meter^3 (Pascal)
                 160.2176565e+9 Pascal = 160.2176565 GPa
                 in SI units and the result will be calculated as follows:
                 output_value = input_value * factor
        """
        newunit = unit.lower()
        newunit = newunit.replace('-','*').replace(' ', '*').replace('^', "**")
        for key,value in unitdict.items():
            newunit = newunit.replace(str(key), str(value))
        newunit = NOTEXCEPT.sub('', newunit)
        try:
            return float(eval(newunit))
        except (ValueError,TypeError):
            return None

    def checkTestsDicts(self, item, localdict):
        for depdict in item["depends"]:
            deptests = depdict["test"]
            depmeet = 0
            for deptest in deptests:
                if deptest[0] in localdict:
                    checkval = localdict[deptest[0]]
                else:
                    accessName, checkval = self.findNameInLookupDict(deptest[0], item.lookupdict)
                    localdict.update({deptest[0] : checkval})
                if(('<' in deptest[1] or   # In Python 3, different type comparisons
                    '>' in deptest[1]) and # are removed. Therefore, < and > comparisons
                    (checkval is None)):   # with a None value generates TypeError
                    pass
                else:
                    if isinstance(checkval, str):
                        if eval('"' + str(checkval) + '"' + deptest[1]):
                            depmeet += 1
                    else:
                        if eval(str(checkval) + deptest[1]):
                            depmeet += 1
                if depmeet == len(deptests):
                    storeValue = False
                    if 'assign' in depdict:
                        return storeValue, depdict['assign'], localdict
                    elif 'value' in depdict:
                        if depdict['value'] in localdict:
                            checkval = localdict[depdict['value']]
                        else:
                            accessName, checkval = self.findNameInLookupDict(depdict['value'],
                                    item.lookupdict)
                            localdict.update({depdict['value'] : checkval})
                        return storeValue, checkval, localdict
                    elif "store" in depdict:
                        itemdepval = depdict["store"]
                        storeValue = True
                        if itemdepval in localdict:
                            checkval = localdict[itemdepval]
                        else:
                            accessName, checkval = self.findNameInLookupDict(itemdepval, item.lookupdict)
                            localdict.update({itemdepval : checkval})
                        print("PRINTING: metainfo, store:",storeValue, checkval)
                        return storeValue, checkval, localdict
        return False, None, localdict

    def checkTestsAttr(self, item, localdict):
        for depdict in item["depends"]:
            #depdict = item["depends"][tests]
            for deptests in depdict["test"]:
                depmeet = 0
                for deptest in deptests:
                    if deptest[0] in localdict:
                        checkval = localdict[deptest[0]]
                    else:
                        attrdict = {deptest[0] : ''}
                        attrdict = self.fetchAttr(attrdict)
                        localdict.update(attrdict)
                        checkval = attrdict[deptest[0]]
                    if eval(str(checkval) + deptest[1]):
                        depmeet += 1
                if depmeet == len(deptests):
                    if 'assign' in depdict:
                        return depdict['assign'], localdict
                    elif 'value' in depdict:
                        if depdict['value'] in localdict:
                            checkval = localdict[depdict['value']]
                        else:
                            attrdict = {depdict['value'] : ''}
                            attrdict = self.fetchAttr(attrdict)
                            localdict.update(attrdict)
                            checkval = attrdict[deptest[0]]
                        return checkval, localdict
        return None, localdict

    def findNameInLookupDict(self, metaname, lookupdict):
        for item in lookupdict:
            itemMap = lookupdict[item]
            #if metaname in itemMap.metaName:
            if metaname in itemMap.matchStr:
                return item, itemMap.value
        return None, None

    def updateBackendStorage(self, backend):
        for itemk in self.Storage.__dict__:
            if(self.Storage.__dict__[itemk]["act"] or
                self.Storage.__dict__[itemk]["stor"]):
                self.Storage.__dict__[itemk]["act"] = False
                self.Storage.__dict__[itemk]["stor"] = False
                value = self.Storage.__dict__[itemk]["val"]
                if isinstance(value, np.ndarray):
                    backend.addArrayValues(itemk, value)
                elif isinstance(value, (list, tuple)):
                    backend.addArrayValues(itemk, np.asarray(value))
                elif value is None:
                    pass
                else:
                    backend.addValue(itemk, value)

    def accumulateDict(self, checkDict):
        localdict = {}
        for itemk in checkDict:
            itemv = checkDict[itemk]
            storeValue, updateValue, localdict = self.checkUpdateValue(itemv, localdict)
            if updateValue is not None:
                itemkinlist = True if itemk in self.Storage.__dict__ else False
                if itemkinlist is False:
                    if (itemk.startswith('x_') and
                        itemv.activeSections):
                        attr_update=False
                        for actSect in itemv.activeSections:
                            if actSect in self.Name:
                                attr_update=True
                        if attr_update:
                            attr_act = False
                            attrvalues = {
                                'act' : attr_act,
                                'val' : None,
                                'stor': False,
                                'kind': None,
                                'dtyp': "C",
                                'unit': None,
                                'size': [],
                                'refs': None
                                }
                            self.Storage.__dict__.update({itemk : attrvalues})
                if itemk in self.Storage.__dict__:
                    if storeValue:
                        #If we need to store the updated values
                        if self.Storage.__dict__[itemk]["val"] is None:
                            #If not initialized, initialize with a list to store
                            if isinstance(updateValue, str) and float(updateValue):
                                self.Storage.__dict__[itemk]["val"] = [float(updateValue)]
                                self.Storage.__dict__[itemk]["stor"] = True
                            else:
                                self.Storage.__dict__[itemk]["val"] = [updateValue]
                                self.Storage.__dict__[itemk]["stor"] = True
                        else:
                            #Append to the stored list if there is a list
                            if type(self.Storage.__dict__[itemk]["val"]) is list:
                                #self.Storage.__dict__[itemk]["val"].append(updateValue)
                                if isinstance(updateValue, str) and float(updateValue):
                                    self.Storage.__dict__[itemk]["val"].append(float(updateValue))
                                    self.Storage.__dict__[itemk]["stor"] = True
                                else:
                                    self.Storage.__dict__[itemk]["val"].append(updateValue)
                                    self.Storage.__dict__[itemk]["stor"] = True
                            else:
                                #Convert the prevoius update to list and append update
                                preValue = self.Storage.__dict__[itemk]["val"]
                                self.Storage.__dict__[itemk]["val"] = [preValue, updateValue]
                                self.Storage.__dict__[itemk]["stor"] = True
                    else:
                        #No need to store, assign the updated value
                        self.Storage.__dict__[itemk]["val"] = updateValue
                    if itemv.activeInfo:
                        if (itemk.startswith('x_') and
                            itemv.activeSections):
                            attr_update=False
                            for actSect in itemv.activeSections:
                                if actSect in self.Name:
                                    attr_update=True
                            if attr_update:
                                self.Storage.__dict__[itemk]["act"] = True
                        else:
                            self.Storage.__dict__[itemk]["act"] = True
                    else:
                        self.Storage.__dict__[itemk]["act"] = False
                self.Active = True
                if "valueSize" in itemv:
                    if "sizeMetaName" in itemv:
                        self.Storage.__dict__[itemv["sizeMetaName"]] = itemv["valueSize"]
                if "unitconverter" in itemv:
                    newValue = itemv["unitconverter"](self, itemv)
                    self.Storage.__dict__[itemk["val"]] = newvalue

    def resetAllValues(self):
        for itemk in self.Storage.__dict__:
            self.Storage.__dict__[itemk]["val"] = None
            self.Storage.__dict__[itemk]["stor"] = False
            self.Storage.__dict__[itemk]["act"] = False
            self.Active = False

    def __str__(self, caller=None, decorate='', color=None, printactive=None, onlynames=None):
        string = ''
        if onlynames is None:
            if self.PrintOnlyNames:
                onlynames = self.PrintOnlyNames
        if printactive is None:
            if self.PrintOnlyActive:
                printactive = self.PrintOnlyActive
        if color:
            color = int(color) + 1
        else:
            if self.Color:
                color = int(self.Color)
        printok = False
        if printactive:
            if self.Active:
                printok = True
        else:
            printok = True
        if caller:
            if printok:
                if color:
                    string = '%s\033[9%sm`-->[' % (decorate + self.Indent, str(color%6 + 1))
                    string = string + ','.join(['%s' % (name) for name in self.Name]) + ']\n\033[0m'
                else:
                    string = '%s`-->[' % (decorate + self.Indent)
                    string = string + ','.join(['%s' % (name) for name in self.Name]) + ']\n'
        else:
            if printok:
                if color:
                    string = '%s\033[9%sm-->[' % (decorate + self.Indent, str(color%6 + 1))
                    string = string + ','.join(['%s' % (name) for name in self.Name]) + ']\n\033[0m'
                else:
                    string = '%s-->[' % (decorate + self.Indent)
                    string = string + ','.join(['%s' % (name) for name in self.Name]) + ']\n'
        if printok:
            if color:
                string = string + '%s\033[9%sm|\033[0m   `.\n' % (decorate + self.Indent, str(color%6 + 1))
            else:
                string = string + '%s|   `.\n' % (decorate + self.Indent)
        if self.Storage:
            for key in self.Storage.__dict__:
                printattrok = False
                if printactive:
                    if self.Storage.__dict__[key]["act"]:
                        printattrok = True
                else:
                    printattrok = True
                if printattrok and printok:
                    if color:
                        if onlynames:
                            string = string + '%s\033[9%sm|\033[0m    |__.%s\n' % (decorate
                                    + self.Indent, str(color%6 + 1), key)
                        else:
                            string = string + '%s\033[9%sm|\033[0m    |__.%s : Active=%s Value=%s\n' % (decorate
                                    + self.Indent, str(color%6 + 1), key, self.Storage.__dict__[key]["act"],
                                    self.Storage.__dict__[key]["val"])
                    else:
                        if onlynames:
                            string = string + '%s|    |__.%s\n' % (decorate +
                                    self.Indent, key)
                        else:
                            string = string + '%s|    |__.%s : Active=%s Value=%s\n' % (decorate +
                                    self.Indent, key, self.Storage.__dict__[key]["act"],
                                    self.Storage.__dict__[key]["val"])
            if color:
                string = string + '%s\033[9%sm|\033[0m\n' % (decorate + self.Indent, str(color%6 + 1))
            else:
                string = string + '%s|\n' % (decorate + self.Indent)
        if self.Containers:
            for module in self.Containers:
                if color:
                    string = string + '%s\033[9%sm|\033[0m\n' % (decorate + self.Indent,
                            str(color%6 + 1)) + module.__str__(self.Name,
                            '%s\033[9%sm|\033[0m' % (decorate + self.Indent, str(color%6 + 1)),
                            color, printactive, onlynames)
                else:
                    string = string + '%s|\n' % (decorate + self.Indent) + module.__str__(self.Name,
                            '%s|' % (decorate + self.Indent),
                            color, printactive, onlynames)
        return string

class Storage(dict):
    """ Storage for meta info document types.
        Sections are build by Container class
    """
    def __init__(self, *args, **kwargs):
        super(Storage, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v

        if kwargs:
            for k, v in kwargs.items():
                self[k] = v

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super(Storage, self).__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super(Storage, self).__delitem__(key)
        del self.__dict__[key]


class JsonMetaInfo(object):
    """ Json file loader for meta info data of NOMAD.
        Loads data and extracts values of items
        with specified superNames
    """
    def __init__(self, metainfo_env):
        self.jsonList = metainfo_env.to_legacy_dict(metainfo_env.packages)['metaInfos']

    def attributes(self, sectionname):
        attributes = {}
        for item in self.jsonList:
            superlist = item['superNames']
            itemname = item['name']
            try:
                kindname = item['kindStr']
            except:
                kindname = []
            try:
                dtyp = item['dtypeStr']
            except:
                dtyp = []
            try:
                size = item['shape']
            except:
                size = []
            try:
                unit = item['units']
            except:
                unit = []
            try:
                refs = item['referencedSections']
            except:
                refs = []
            if ('type_section' in kindname or
                sectionname not in superlist):
                continue
            attrvalues = {
                    'act' : False,
                    'stor': False,
                    'val' : None,
                    'kind': kindname,
                    'dtyp': dtyp,
                    'unit': unit,
                    'size': size,
                    'refs': refs
                    }
            attributes.update({itemname: attrvalues})
        return attributes

    def siblings(self, sectionname):
        siblings = []
        searchList = []
        nameList = []
        containsList = []
        for item in self.jsonList:
            superlist = item['superNames']
            itemname = item['name']
            try:
                kindname = item['kindStr']
            except:
                kindname = []
            if ('type_section' in kindname or
                'type_abstract_document_content' in kindname):
                if sectionname in superlist:
                    searchList.append(itemname)
                if itemname not in nameList:
                    nameList.append(itemname)
        for name in searchList:
            if (set([name]) not in set(nameList) and
                self.isparent(name)):
                siblings.append(name)
        return siblings

    def isparent(self, itemname):
        haschild = False
        for item in self.jsonList:
            if itemname in item['superNames']:
                haschild = True
        return haschild

    def rootsections(self):
        rootname = []
        searchList = []
        nameList = []
        for item in self.jsonList:
            superlist = item['superNames']
            itemname = item['name']
            try:
                kindname = item['kindStr']
            except:
                kindname = []
            if 'type_section' not in kindname:
                continue
            if not superlist:
                searchList.append(itemname)
            if itemname not in nameList:
               nameList.append(itemname)
        for name in searchList:
            if ('section' in name and set([name]) not in set(nameList)):
                rootname.append(name)
        return rootname

    def fetchdict(self, itemname, pattern):
        resDict = {}
        for item in self.jsonList:
            val = dict(item)
            itemProperty = item[itemname]
            if pattern in itemProperty:
                resDict.update({item["name"]: val})
        return resDict

if __name__ == "__main__":
    run = Container('section_run')
    exclude_dict = {
            'section_run' : [
            'section_processor_info',
            'section_processor_log',
            'section_springer_material',
            'section_repository_info'
            ]}

    from nomad.datamodel.metainfo import m_env
    jsonmetadata = JsonMetaInfo(m_env)

    updateDict = {
            'startSection' : [['section_topology']],
            'muteSections' : [['section_interaction']],
            'dictionary' : {
                'molecule_constraint_atoms' : {'depends' : [[]], 'assign' : 100},
                'interaction_atoms' : {'depends' : [[]], 'assign' : 10},
                'topology_force_field_name' : {'depends' : [[]], 'assign' : "ReaxFF"}
                }
            }

    run.populate(jsonmetadata, 'section_run', exclude_dict, updateDict)
    run.Color = 4
    for container in run.Containers:
        if 'section_topology' in container.Name:
            select = container
    select.Color = 4
    #select.PrintOnlyActive = 1
    run.PrintOnlyNames = 1
    print(run)





# Copyright 2015-2018 Martina Stella, Massimo Riello, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import division
from builtins import range
from builtins import object
import numpy as np
import nomadcore.ActivateLogging
from nomadcore.caching_backend import CachingLevel
from nomadcore.simple_parser import mainFunction
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from castepparser.CastepCommon import get_metaInfo
import logging, os, re, sys




###############################################################
# This is the parser for the *.band file of CASTEP.


# NB: this parser store self consistent eigenvalues and
#     relative k points for single configuration calculations

###############################################################

logger = logging.getLogger("nomad.CastepBandParser")

class CastepBandParserContext(object):
    """Context for parsing CASTEP *.band file.


    The onClose_ functions allow processing and writing of cached values after a section is closed.
    They take the following arguments:
        backend: Class that takes care of wrting and caching of metadata.
        gIndex: Index of the section that is closed.
        section: The cached values and sections that were found in the section that is closed.
    """
    def __init__(self, writeMetaData = True):
        """Args:
            writeMetaData: Deteremines if metadata is written or stored in class attributes.
        """
        self.writeMetaData = writeMetaData

    def startedParsing(self, fInName, parser):
        """Function is called when the parsing starts and the compiled parser is obtained.

        Args:
            fInName: The file name on which the current parser is running.
            parser: The compiled parser. Is an object of the class SimpleParser in nomadcore.simple_parser.py.
        """
        self.parser = parser
        # get unit from metadata for band energies
        # allows to reset values if the same superContext is used to parse different files
        self.k_count                    = 0
        self.k_nr                       = 0
        self.e_nr                       = 0
        self.eigenvalues_kpoints        = []
        self.eigenvalues_values         = []

        self.e_spin_1                   = []
        self.e_spin_2                   = []
        self.n_spin                     = 0



# Reading the number of spins
    def onClose_x_castep_section_spin_number(self, backend, gIndex, section):

        self.n_spin = section['x_castep_spin_number']


# Storing the k point coordinates
    def onClose_x_castep_section_scf_k_points(self, backend, gIndex, section):
        """trigger called when _section_eigenvalues"""

# Processing k points (given in fractional coordinates)
        #get cached values of x_castep_store_k_points
        k_st = section['x_castep_store_scf_k_points']
        self.k_count = len(k_st)
        self.k_nr   += 1
        for i in range(0, self.k_count):
            k_st[i] = k_st[i].split()
            k_st[i] = [float(j) for j in k_st[i]]
            k_st_int = k_st[i]
            self.eigenvalues_kpoints.append(k_st_int)


# Storing the eigenvalues
    def onClose_x_castep_section_scf_eigenvalues(self, backend, gIndex, section):
        """trigger called when _section_eigenvalues"""
        Ha_to_J = 4.35974e-18
        #get cached values of castep_store_k_points
        e_st = section['x_castep_store_scf_eigenvalues']

        e_st_0 = e_st
        e_st_0 = [x * Ha_to_J for x in e_st_0]

        def split_list(lista):
            half = len(lista) // 2
            return lista[:half], lista[half:]

        e_st_1, e_st_2 = split_list(e_st)
        e_st_1 = [x * Ha_to_J for x in e_st_1]
        e_st_2 = [x * Ha_to_J for x in e_st_2]

        if self.n_spin[0] == 1:
            self.e_nr = len(e_st_0)
            self.e_spin_1.append(e_st_0)
            self.e_spin_2 = []

        else:
            self.e_nr = len(e_st_1)
            self.e_spin_1.append(e_st_1)
            self.e_spin_2.append(e_st_2)







################################################################################
######################  MAIN PARSER STARTS HERE  ###############################
################################################################################


def build_CastepBandFileSimpleMatcher():
    """Builds the SimpleMatcher to parse the *.cell file of CASTEP.

    SimpleMatchers are called with 'SM (' as this string has length 4,
    which allows nice formating of nested SimpleMatchers in python.

    Returns:
       SimpleMatcher that parses *.cell file of CASTEP.
    """
    return SM (name = 'Root1',
        startReStr = "",
        sections = ['section_run'],
        forwardMatch = True,
        weak = True,
        subMatchers = [
        SM (name = 'Root2',
            startReStr = "",
            sections = ['section_single_configuration_calculation'],
            forwardMatch = True,
            weak = True,
            subMatchers = [

               SM(startReStr = r"Number\sof\sk\-points\s*[0-9]+\s*",
                  sections = ['x_castep_section_spin_number'],
                  forwardMatch = True,
                  subMatchers = [

                     SM(r"Number\sof\sspin\scomponents\s*(?P<x_castep_spin_number>[1-2]+)")

                                 ]),

               SM(startReStr = r"K\-point\s*[0-9]+\s*",
                  sections = ["x_castep_section_scf_k_points"],
                  forwardMatch = True,
                  repeats = True,
                  subMatchers = [

                     SM(r"K\-point\s*[0-9]+\s*(?P<x_castep_store_scf_k_points> [-\d\.]+\s+[-\d\.]+\s+[-\d\.]+)",
                        repeats = True),

                     SM(name = 'Eigen',
                        startReStr = r"Spin component\s*1\s*",
                        #endReStr = r"Spin component\s*2\s*",
                        sections = ['x_castep_section_scf_eigenvalues'],
                        repeats = True,
                        subMatchers = [

                           SM(r"\s*(?P<x_castep_store_scf_eigenvalues> [-\d\.]+)",
                              repeats = True)

                                       ]), # CLOSING castep_section_scf_eigenvalues


                                 ]) # CLOSING castep_section_k_points


    ]), # CLOSING section_single_configuration_calculation
])


def get_cachingLevelForMetaName(metaInfoEnv, CachingLvl):
    """Sets the caching level for the metadata.

    Args:
        metaInfoEnv: metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
        CachingLvl: Sets the CachingLevel for the sections k_band, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.

    Returns:
        Dictionary with metaname as key and caching level as value.
    """
    # manually adjust caching of metadata
    cachingLevelForMetaName = {
                               'section_run': CachingLvl,
                               'section_single_configuration_calculation': CachingLvl,
                              }
    # Set all band metadata to Cache as they need post-processsing.
    for name in metaInfoEnv.infoKinds:
        if name.startswith('x_castep_'):
            cachingLevelForMetaName[name] = CachingLevel.Cache
    return cachingLevelForMetaName


def main(CachingLvl):
    """Main function.

    Set up everything for the parsing of the CASTEP *.cell file and run the parsing.

    Args:
        CachingLvl: Sets the CachingLevel for the sections k_band, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.
    """
    # get band.out file description
    CasteBandFileSimpleMatcher = build_CastepBandFileSimpleMatcher()
    # loading metadata from nomad-meta-info/meta_info/nomad_meta_info/castep.nomadmetainfo.json
    metaInfoPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../../../../nomad-meta-info/meta_info/nomad_meta_info/castep.nomadmetainfo.json"))
    metaInfoEnv = get_metaInfo(metaInfoPath)
    # set parser info
    parserInfo = {'name':'castep-cell-parser', 'version': '1.0'}
    # get caching level for metadata
    cachingLevelForMetaName = get_cachingLevelForMetaName(metaInfoEnv, CachingLvl)
    # start parsing
    mainFunction(mainFileDescription = CasteBandFileSimpleMatcher,
                 metaInfoEnv = metaInfoEnv,
                 parserInfo = parserInfo,
                 cachingLevelForMetaName = cachingLevelForMetaName,
                 superContext = CastepBandParserContext())

if __name__ == "__main__":
    main(CachingLevel.Forward)


